#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"Define all legal items and values for cylc suite definition files."

import re

from parsec.validate import validator as vdr
from parsec.validate import (
    coercers, _strip_and_unquote, _strip_and_unquote_list, IllegalValueError)
from parsec.util import itemstr
from parsec.upgrade import upgrader, converter
from parsec.config import config
from cylc.syntax_flags import (
    set_syntax_version, VERSION_PREV, VERSION_NEW, SyntaxVersionError)
from isodatetime.dumpers import TimePointDumper
from isodatetime.data import Calendar, TimePoint
from isodatetime.parsers import TimePointParser, DurationParser
from cylc.cycling.integer import REC_INTERVAL as REC_INTEGER_INTERVAL

from cylc.cfgspec.utils import (
    coerce_interval, coerce_interval_list, DurationFloat)
from cylc.cfgspec.globalcfg import GLOBAL_CFG
from cylc.network import PRIVILEGE_LEVELS


REC_PARAM_INT_RANGE = re.compile('(\d+)\.\.(\d+)')


def _coerce_cycleinterval(value, keys, _):
    """Coerce value to a cycle interval."""
    if not value:
        return None
    value = _strip_and_unquote(keys, value)
    if value.isdigit():
        # Old runahead limit format.
        set_syntax_version(VERSION_PREV,
                           "integer interval for %s" % itemstr(
                               keys[:-1], keys[-1], value))
        return value
    if REC_INTEGER_INTERVAL.match(value):
        # New integer cycling format.
        set_syntax_version(VERSION_NEW,
                           "integer interval for %s" % itemstr(
                               keys[:-1], keys[-1], value))
        return value
    parser = DurationParser()
    try:
        parser.parse(value)
    except ValueError:
        raise IllegalValueError("interval", keys, value)
    set_syntax_version(VERSION_NEW,
                       "ISO 8601 interval for %s" % itemstr(
                           keys[:-1], keys[-1], value))
    return value


def _coerce_cycletime(value, keys, _):
    """Coerce value to a cycle point."""
    if not value:
        return None
    if value == "now":
        # Handle this later in config.py when the suite UTC mode is known.
        return value
    value = _strip_and_unquote(keys, value)
    if re.match(r"\d+$", value):
        # Could be an old date-time cycle point format, or integer format.
        return value
    if value.startswith("-") or value.startswith("+"):
        # We don't know the value given for num expanded year digits...
        for i in range(1, 101):
            parser = TimePointParser(num_expanded_year_digits=i)
            try:
                parser.parse(value)
            except ValueError:
                continue
            return value
        raise IllegalValueError("cycle point", keys, value)
    parser = TimePointParser()
    try:
        parser.parse(value)
    except ValueError:
        raise IllegalValueError("cycle point", keys, value)
    set_syntax_version(VERSION_NEW,
                       "cycle point: %s" % itemstr(
                           keys[:-1], keys[-1], value))
    return value


def _coerce_cycletime_format(value, keys, _):
    """Coerce value to a cycle point format (either CCYYMM... or %Y%m...)."""
    value = _strip_and_unquote(keys, value)
    if not value:
        return None
    try:
        set_syntax_version(VERSION_NEW, "use of [cylc]cycle point format")
    except SyntaxVersionError:
        raise IllegalValueError("cycle point format", keys, value)
    test_timepoint = TimePoint(year=2001, month_of_year=3, day_of_month=1,
                               hour_of_day=4, minute_of_hour=30,
                               second_of_minute=54)
    if "/" in value:
        raise IllegalValueError("cycle point format", keys, value)
    if "%" in value:
        try:
            TimePointDumper().strftime(test_timepoint, value)
        except ValueError:
            raise IllegalValueError("cycle point format", keys, value)
        return value
    if "X" in value:
        for i in range(1, 101):
            dumper = TimePointDumper(num_expanded_year_digits=i)
            try:
                dumper.dump(test_timepoint, value)
            except ValueError:
                continue
            return value
        raise IllegalValueError("cycle point format", keys, value)
    dumper = TimePointDumper()
    try:
        dumper.dump(test_timepoint, value)
    except ValueError:
        raise IllegalValueError("cycle point format", keys, value)
    return value


def _coerce_cycletime_time_zone(value, keys, _):
    """Coerce value to a cycle point time zone format - Z, +13, -0800..."""
    value = _strip_and_unquote(keys, value)
    if not value:
        return None
    try:
        set_syntax_version(
            VERSION_NEW, "use of [cylc]cycle point time zone format")
    except SyntaxVersionError:
        raise IllegalValueError("cycle point time zone format", keys, value)
    test_timepoint = TimePoint(year=2001, month_of_year=3, day_of_month=1,
                               hour_of_day=4, minute_of_hour=30,
                               second_of_minute=54)
    dumper = TimePointDumper()
    test_timepoint_string = dumper.dump(test_timepoint, "CCYYMMDDThhmmss")
    test_timepoint_string += value
    parser = TimePointParser(allow_only_basic=True)
    try:
        parser.parse(test_timepoint_string)
    except ValueError:
        raise IllegalValueError("cycle point time zone format", keys, value)
    return value


def _coerce_final_cycletime(value, keys, _):
    """Coerce final cycle point."""
    return _strip_and_unquote(keys, value)


def _coerce_parameter_list(value, keys, _):
    """Coerce parameter list."""
    value = _strip_and_unquote_list(keys, value)
    if len(value) == 1:
        # May be a range e.g. '1..5' (bounds inclusive)
        try:
            lower, upper = REC_PARAM_INT_RANGE.match(value[0]).groups()
        except AttributeError:
            if '.' in value[0]:
                # Dot is illegal in node names, probably bad range syntax.
                raise IllegalValueError("parameter", keys, value)
        else:
            n_dig = len(upper)
            return [
                str(i).zfill(n_dig) for i in range(int(lower), int(upper) + 1)]
    return value

coercers['cycletime'] = _coerce_cycletime
coercers['cycletime_format'] = _coerce_cycletime_format
coercers['cycletime_time_zone'] = _coerce_cycletime_time_zone
coercers['cycleinterval'] = _coerce_cycleinterval
coercers['final_cycletime'] = _coerce_final_cycletime
coercers['interval'] = coerce_interval
coercers['interval_minutes'] = lambda *a: coerce_interval(
    *a, back_comp_unit_factor=60)
coercers['interval_seconds'] = coerce_interval
coercers['interval_list'] = coerce_interval_list
coercers['interval_minutes_list'] = lambda *a: coerce_interval_list(
    *a, back_comp_unit_factor=60)
coercers['interval_seconds_list'] = coerce_interval_list
coercers['parameter_list'] = _coerce_parameter_list


SPEC = {
    'title': vdr(vtype='string', default=""),
    'description': vdr(vtype='string', default=""),
    'URL': vdr(vtype='string', default=""),
    'cylc': {
        'UTC mode': vdr(
            vtype='boolean', default=GLOBAL_CFG.get(['cylc', 'UTC mode'])),
        'cycle point format': vdr(
            vtype='cycletime_format', default=None),
        'cycle point num expanded year digits': vdr(
            vtype='integer', default=0),
        'cycle point time zone': vdr(
            vtype='cycletime_time_zone', default=None),
        'required run mode': vdr(
            vtype='string', options=['live', 'dummy', 'simulation', '']),
        'force run mode': vdr(
            vtype='string', options=['live', 'dummy', 'simulation', '']),
        'abort if any task fails': vdr(vtype='boolean', default=False),
        'log resolved dependencies': vdr(vtype='boolean', default=False),
        'disable automatic shutdown': vdr(vtype='boolean', default=False),
        'environment': {
            '__MANY__': vdr(vtype='string'),
        },
        'parameters': {
            '__MANY__': vdr(vtype='parameter_list'),
        },
        'parameter templates': {
            '__MANY__': vdr(vtype='string'),
        },
        'events': {
            'handlers': vdr(vtype='string_list'),
            'handler events': vdr(vtype='string_list'),
            'startup handler': vdr(vtype='string_list'),
            'timeout handler': vdr(vtype='string_list'),
            'inactivity handler': vdr(vtype='string_list'),
            'shutdown handler': vdr(vtype='string_list'),
            'stalled handler': vdr(vtype='string_list'),
            'timeout': vdr(vtype='interval_minutes'),
            'inactivity': vdr(vtype='interval_minutes'),
            'reset timer': vdr(vtype='boolean', default=True),
            'reset inactivity timer': vdr(vtype='boolean', default=True),
            'abort if startup handler fails': vdr(
                vtype='boolean', default=False),
            'abort if shutdown handler fails': vdr(
                vtype='boolean', default=False),
            'abort if timeout handler fails': vdr(
                vtype='boolean', default=False),
            'abort if inactivity handler fails': vdr(
                vtype='boolean', default=False),
            'abort if stalled handler fails': vdr(
                vtype='boolean', default=False),
            'abort on stalled': vdr(vtype='boolean', default=None),
            'abort on timeout': vdr(vtype='boolean', default=None),
            'abort on inactivity': vdr(vtype='boolean'),
            'mail events': vdr(vtype='string_list'),
            'mail from': vdr(vtype='string'),
            'mail smtp': vdr(vtype='string'),
            'mail to': vdr(vtype='string'),
            'mail footer': vdr(vtype='string'),
        },
        'simulation mode': {
            'disable suite event hooks': vdr(vtype='boolean', default=True),
        },
        'dummy mode': {
            'disable suite event hooks': vdr(vtype='boolean', default=True),
        },
        'reference test': {
            'suite shutdown event handler': vdr(
                vtype='string', default='cylc hook check-triggering'),
            'required run mode': vdr(
                vtype='string', options=['live', 'simulation', 'dummy', '']),
            'allow task failures': vdr(vtype='boolean', default=False),
            'expected task failures': vdr(vtype='string_list', default=[]),
            'live mode suite timeout': vdr(
                vtype='interval_minutes', default=DurationFloat(60)),
            'dummy mode suite timeout': vdr(
                vtype='interval_minutes', default=DurationFloat(60)),
            'simulation mode suite timeout': vdr(
                vtype='interval_minutes', default=DurationFloat(60)),
        },
        'authentication': {
            # Allow owners to grant public shutdown rights at the most, not
            # full control.
            'public': vdr(
                vtype='string',
                options=PRIVILEGE_LEVELS[
                    :PRIVILEGE_LEVELS.index('shutdown') + 1],
                default=GLOBAL_CFG.get(['authentication', 'public']))
        },
    },
    'scheduling': {
        'initial cycle point': vdr(vtype='cycletime'),
        'final cycle point': vdr(vtype='final_cycletime'),
        'initial cycle point constraints': vdr(
            vtype='string_list', default=[]),
        'final cycle point constraints': vdr(vtype='string_list', default=[]),
        'hold after point': vdr(vtype='cycletime'),
        'cycling mode': vdr(
            vtype='string',
            default=Calendar.MODE_GREGORIAN,
            options=(Calendar.MODES.keys() + ["integer"])),
        'runahead limit': vdr(vtype='cycleinterval'),
        'max active cycle points': vdr(vtype='integer', default=3),
        'spawn to max active cycle points': vdr(
            vtype='boolean', default=False),
        'queues': {
            'default': {
                'limit': vdr(vtype='integer', default=0),
                'members': vdr(vtype='string_list', default=[]),
            },
            '__MANY__': {
                'limit': vdr(vtype='integer', default=0),
                'members': vdr(vtype='string_list', default=[]),
            },
        },
        'special tasks': {
            'clock-trigger': vdr(vtype='string_list', default=[]),
            'external-trigger': vdr(vtype='string_list', default=[]),
            'clock-expire': vdr(vtype='string_list', default=[]),
            'sequential': vdr(vtype='string_list', default=[]),
            'start-up': vdr(vtype='string_list', default=[]),
            'cold-start': vdr(vtype='string_list', default=[]),
            'exclude at start-up': vdr(vtype='string_list', default=[]),
            'include at start-up': vdr(vtype='string_list', default=[]),
        },
        'dependencies': {
            'graph': vdr(vtype='string'),
            '__MANY__':
            {
                'graph': vdr(vtype='string'),
            },
        },
    },
    'runtime': {
        '__MANY__': {
            'inherit': vdr(vtype='string_list', default=[]),
            'title': vdr(vtype='string', default=""),
            'description': vdr(vtype='string', default=""),
            'URL': vdr(vtype='string', default=""),
            'init-script': vdr(vtype='string'),
            'env-script': vdr(vtype='string'),
            'pre-script': vdr(vtype='string'),
            'script': vdr(
                vtype='string',
                default='echo Dummy task; sleep $(cylc rnd 1 16)'),
            'post-script': vdr(vtype='string'),
            'extra log files': vdr(vtype='string_list', default=[]),
            'enable resurrection': vdr(vtype='boolean', default=False),
            'work sub-directory': vdr(
                vtype='string',
                default='$CYLC_TASK_CYCLE_POINT/$CYLC_TASK_NAME'),
            'environment filter': {
                'include': vdr(vtype='string_list'),
                'exclude': vdr(vtype='string_list'),
            },
            'simulation mode': {
                'run time range': vdr(
                    vtype='interval_seconds_list',
                    default=[DurationFloat(1), DurationFloat(16)]),
                'simulate failure': vdr(vtype='boolean', default=False),
                'disable task event hooks': vdr(vtype='boolean', default=True),
                'disable retries': vdr(vtype='boolean', default=True),
            },
            'dummy mode': {
                'script': vdr(
                    vtype='string',
                    default='echo Dummy task; sleep $(cylc rnd 1 16)'),
                'disable pre-script': vdr(vtype='boolean', default=True),
                'disable post-script': vdr(vtype='boolean', default=True),
                'disable task event hooks': vdr(vtype='boolean', default=True),
                'disable retries': vdr(vtype='boolean', default=True),
            },
            'job': {
                'batch system': vdr(vtype='string', default='background'),
                'batch submit command template': vdr(vtype='string'),
                'execution polling intervals': vdr(
                    vtype='interval_minutes_list', default=[]),
                'execution retry delays': vdr(
                    vtype='interval_minutes_list', default=[]),
                'execution time limit': vdr(vtype='interval_seconds'),
                'shell': vdr(vtype='string', default='/bin/bash'),
                'submission polling intervals': vdr(
                    vtype='interval_minutes_list', default=[]),
                'submission retry delays': vdr(
                    vtype='interval_minutes_list', default=[]),
            },
            'remote': {
                'host': vdr(vtype='string'),
                'owner': vdr(vtype='string'),
                'suite definition directory': vdr(vtype='string'),
                'retrieve job logs': vdr(vtype='boolean', default=None),
                'retrieve job logs max size': vdr(vtype='string'),
                'retrieve job logs retry delays': vdr(
                    vtype='interval_minutes_list'),
            },
            'events': {
                'execution timeout': vdr(vtype='interval_minutes'),
                'handlers': vdr(vtype='string_list'),
                'handler events': vdr(vtype='string_list'),
                'handler retry delays': vdr(vtype='interval_minutes_list'),
                'mail events': vdr(vtype='string_list'),
                'mail from': vdr(vtype='string'),
                'mail retry delays': vdr(vtype='interval_minutes_list'),
                'mail smtp': vdr(vtype='string'),
                'mail to': vdr(vtype='string'),
                'reset timer': vdr(vtype='boolean', default=None),
                'submission timeout': vdr(vtype='interval_minutes'),

                'expired handler': vdr(vtype='string_list', default=[]),
                'submitted handler': vdr(vtype='string_list', default=[]),
                'started handler': vdr(vtype='string_list', default=[]),
                'succeeded handler': vdr(vtype='string_list', default=[]),
                'failed handler': vdr(vtype='string_list', default=[]),
                'submission failed handler': vdr(
                    vtype='string_list', default=[]),
                'warning handler': vdr(vtype='string_list', default=[]),
                'retry handler': vdr(vtype='string_list', default=[]),
                'submission retry handler': vdr(
                    vtype='string_list', default=[]),
                'execution timeout handler': vdr(
                    vtype='string_list', default=[]),
                'submission timeout handler': vdr(
                    vtype='string_list', default=[]),
            },
            'suite state polling': {
                'user': vdr(vtype='string'),
                'host': vdr(vtype='string'),
                'interval': vdr(vtype='interval_seconds'),
                'max-polls': vdr(vtype='integer'),
                'run-dir': vdr(vtype='string'),
                'template': vdr(vtype='string'),
                'verbose mode': vdr(vtype='boolean', default=None),
            },
            'environment': {
                '__MANY__': vdr(vtype='string'),
            },
            'directives': {
                '__MANY__': vdr(vtype='string'),
            },
            'outputs': {
                '__MANY__': vdr(vtype='string'),
            },
        },
    },
    'visualization': {
        'initial cycle point': vdr(vtype='cycletime'),
        'final cycle point': vdr(vtype='final_cycletime'),
        'number of cycle points': vdr(vtype='integer', default=3),
        'collapsed families': vdr(vtype='string_list', default=[]),
        'use node color for edges': vdr(vtype='boolean', default=False),
        'use node fillcolor for edges': vdr(vtype='boolean', default=False),
        'use node color for labels': vdr(vtype='boolean', default=False),
        'node penwidth': vdr(vtype='integer', default=2),
        'edge penwidth': vdr(vtype='integer', default=2),
        'default node attributes': vdr(
            vtype='string_list',
            default=['style=unfilled', 'color=black', 'shape=box']),
        'default edge attributes': vdr(
            vtype='string_list', default=['color=black']),
        'node groups': {
            '__MANY__': vdr(vtype='string_list', default=[]),
        },
        'node attributes': {
            '__MANY__': vdr(vtype='string_list', default=[]),
        },
    },
}


def upg(cfg, descr):
    """Upgrade old suite configuration."""
    u = upgrader(cfg, descr)
    u.deprecate(
        '5.2.0',
        ['cylc', 'event handler execution'],
        ['cylc', 'event handler submission'])
    # TODO - should abort if obsoleted items are encountered
    u.obsolete(
        '5.4.7', ['scheduling', 'special tasks', 'explicit restart outputs'])
    u.obsolete('5.4.11', ['cylc', 'accelerated clock'])
    u.obsolete('6.0.0', ['visualization', 'runtime graph'])
    u.obsolete('6.1.3', ['visualization', 'enable live graph movie'])
    u.obsolete('6.0.0', ['development'])
    u.deprecate(
        '6.0.0',
        ['scheduling', 'initial cycle time'],
        ['scheduling', 'initial cycle point'],
        converter(
            lambda x: x, 'changed naming to reflect non-date-time cycling'))
    u.deprecate(
        '6.0.0',
        ['scheduling', 'final cycle time'],
        ['scheduling', 'final cycle point'],
        converter(
            lambda x: x, 'changed naming to reflect non-date-time cycling'))
    u.deprecate(
        '6.0.0',
        ['visualization', 'initial cycle time'],
        ['visualization', 'initial cycle point'],
        converter(
            lambda x: x, 'changed naming to reflect non-date-time cycling'))
    u.deprecate(
        '6.0.0',
        ['visualization', 'final cycle time'],
        ['visualization', 'final cycle point'],
        converter(
            lambda x: x, 'changed naming to reflect non-date-time cycling'))
    u.deprecate(
        '6.0.0',
        ['scheduling', 'cycling']
    )
    u.deprecate(
        '6.0.0',
        ['scheduling', 'special tasks', 'sequential']
    )
    u.deprecate(
        '6.0.0',
        ['scheduling', 'special tasks', 'start-up']
    )
    u.deprecate(
        '6.0.0',
        ['scheduling', 'special tasks', 'cold-start']
    )
    u.obsolete('6.0.0', ['cylc', 'job submission'])
    u.obsolete('6.0.0', ['cylc', 'event handler submission'])
    u.obsolete('6.0.0', ['cylc', 'poll and kill command submission'])
    u.obsolete('6.0.0', ['cylc', 'lockserver'])
    dep = {
        'pre-command scripting': 'pre-script',
        'command scripting': 'script',
        'post-command scripting': 'post-script',
        'environment scripting': 'env-script',
        'initial scripting': 'init-script'
    }
    for old, new in dep.items():
        u.deprecate(
            '6.4.0',
            ['runtime', '__MANY__', old],
            ['runtime', '__MANY__', new],
            silent=True)
        u.deprecate(
            '6.4.0',
            ['runtime', '__MANY__', 'dummy mode', old],
            ['runtime', '__MANY__', 'dummy mode', new],
            silent=True)
    u.deprecate(
        '6.5.0',
        ['scheduling', 'special tasks', 'clock-triggered'],
        ['scheduling', 'special tasks', 'clock-trigger'],
    )
    u.deprecate(
        '6.5.0',
        ['scheduling', 'special tasks', 'external-triggered'],
        ['scheduling', 'special tasks', 'external-trigger'],
    )
    for key in SPEC['cylc']['events']:
        u.deprecate(
            '6.11.0', ['cylc', 'event hooks', key], ['cylc', 'events', key])
    u.deprecate('6.11.0', ['cylc', 'event hooks'])
    for key in SPEC['runtime']['__MANY__']['events']:
        u.deprecate(
            '6.11.0',
            ['runtime', '__MANY__', 'event hooks', key],
            ['runtime', '__MANY__', 'events', key])
    u.deprecate('6.11.0', ['runtime', '__MANY__', 'event hooks'])
    u.deprecate(
        '6.11.0',
        ['runtime', '__MANY__', 'job submission', 'method'],
        ['runtime', '__MANY__', 'job', 'batch system'])
    u.deprecate(
        '6.11.0',
        ['runtime', '__MANY__', 'job submission', 'command template'],
        ['runtime', '__MANY__', 'job', 'batch submit command template'])
    u.deprecate(
        '6.11.0',
        ['runtime', '__MANY__', 'job submission', 'shell'],
        ['runtime', '__MANY__', 'job', 'shell'])
    u.deprecate(
        '6.11.0',
        ['runtime', '__MANY__', 'job submission', 'retry delays'],
        ['runtime', '__MANY__', 'job', 'submission retry delays'])
    u.deprecate('6.11.0', ['runtime', '__MANY__', 'job submission'])
    u.deprecate(
        '6.11.0',
        ['runtime', '__MANY__', 'retry delays'],
        ['runtime', '__MANY__', 'job', 'execution retry delays'])
    u.deprecate(
        '6.11.0',
        ['runtime', '__MANY__', 'submission polling intervals'],
        ['runtime', '__MANY__', 'job', 'submission polling intervals'])
    u.deprecate(
        '6.11.0',
        ['runtime', '__MANY__', 'execution polling intervals'],
        ['runtime', '__MANY__', 'job', 'execution polling intervals'])
    u.upgrade()
    if 'cylc' in cfg and 'event hooks' in cfg['cylc']:
        del cfg['cylc']['event hooks']
    if 'runtime' in cfg:
        for section in cfg['runtime'].values():
            for key in ['event hooks', 'job submission']:
                if key in section:
                    del section[key]

    # Force pre cylc-6 "cycling = Yearly" type suites to the explicit
    # dependency heading form for which backward compatibility is provided:
    # ___________________________
    # [scheduling]
    #    cycling = Yearly
    #    [[dependencies]]
    #        [[[2014,2]]]
    # ---------------------------
    # Same as (for auto upgrade):
    # ---------------------------
    # [scheduling]
    #    [[dependencies]]
    #        [[[Yearly(2014,2)]]]
    # ___________________________
    try:
        old_cycling_mode = cfg['scheduling']['cycling']
    except KeyError:
        pass
    else:
        if old_cycling_mode in ['Yearly', 'Monthly', 'Daily']:
            del cfg['scheduling']['cycling']
            for old_key, val in cfg['scheduling']['dependencies'].items():
                if re.match('\s*\d+,\s*\d+\s*$', old_key):
                    new_key = "%s(%s)" % (old_cycling_mode, old_key)
                    del cfg['scheduling']['dependencies'][old_key]
                    cfg['scheduling']['dependencies'][new_key] = val
        else:
            # Could be misspelled new "cycling mode" - leave it to fail.
            pass


class RawSuiteConfig(config):
    """Raw suite configuration."""
    _SUITECFG = None
    _CFPATH = None

    @classmethod
    def get_inst(cls, fpath, force=False, tvars=None, write_proc=False):
        """Return the default instance."""
        if cls._SUITECFG is None or fpath != cls._CFPATH or force:
            cls._CFPATH = fpath
            if tvars is None:
                tvars = []
            # TODO - write_proc should be in loadcfg
            cls._SUITECFG = cls(SPEC, upg, tvars=tvars, write_proc=write_proc)
            cls._SUITECFG.loadcfg(fpath, "suite definition")
        return cls._SUITECFG



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from Queue import Queue
from cylc.owner import USER
from cylc.suite_host import get_hostname
from cylc.network.pyro_base import PyroClient, PyroServer
from cylc.network import check_access_priv


class TaskMessageServer(PyroServer):
    """Server-side task messaging interface"""

    def __init__(self):
        super(TaskMessageServer, self).__init__()
        self.queue = Queue()

    def put(self, task_id, priority, message):
        check_access_priv(self, 'full-control')
        self.report('task_message')
        self.queue.put((task_id, priority, message))
        return (True, 'Message queued')

    def get_queue(self):
        return self.queue


class TaskMessageClient(PyroClient):
    """Client-side task messaging interface"""

    def __init__(self, suite, task_id, owner=USER,
                 host=get_hostname(), pyro_timeout=None, port=None):
        self.target_server_object = "task_pool"
        self.task_id = task_id
        super(TaskMessageClient, self).__init__(
            suite, owner, host, pyro_timeout, port)

    def put(self, *args):
        args = [self.task_id] + list(args)
        self.call_server_func('put', *args)



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Utility for "cylc cat-state" and "cylc dump"."""

import time
from cylc.task_id import TaskID
from cylc.network.suite_state import SUITE_STATUS_STOPPED
from cylc.task_state import TASK_STATUS_READY


def get_stop_state_summary(lines):
    """Parse state dump content into summary maps."""
    global_summary = {}
    task_summary = {}
    if len(lines) == 0 or len(lines) < 3:
        return None
    for line in list(lines):
        if line.startswith('Remote command'):
            lines.remove(line)
    line0 = lines.pop(0)
    if line0.startswith('suite time') or \
            line0.startswith('simulation time'):
        # backward compatibility with pre-5.4.11 state dumps
        global_summary["last_updated"] = time.time()
    else:
        # (line0 is run mode)
        line1 = lines.pop(0)
        while not line1.startswith("time :"):
            line1 = lines.pop(0)
        try:
            time_string = line1.rstrip().split(' : ')[1]
            unix_time_string = time_string.rsplit('(', 1)[1].rstrip(")")
            global_summary["last_updated"] = int(unix_time_string)
        except (TypeError, ValueError, IndexError):
            # back compat pre cylc-6
            global_summary["last_updated"] = time.time()

    # Skip initial and final cycle points.
    lines[0:2] = []
    global_summary["status_string"] = SUITE_STATUS_STOPPED
    while lines:
        line = lines.pop(0)
        if line.startswith("class") or line.startswith("Begin task"):
            continue
        try:
            (task_id, info) = line.split(' : ')
            name, point_string = TaskID.split(task_id)
        except ValueError:
            continue
        task_summary.setdefault(task_id, {"name": name, "point": point_string,
                                          "label": point_string})
        # reconstruct state from a dumped state string
        items = dict([p.split("=") for p in info.split(', ')])
        state = items.get("status")
        if state == 'submitting':
            # backward compabitility for state dumps generated prior to #787
            state = TASK_STATUS_READY
        task_summary[task_id].update({"state": state})
        task_summary[task_id].update({"spawned": items.get("spawned")})
    global_summary["run_mode"] = "dead"
    return global_summary, task_summary


def dump_to_stdout(states, sort_by_cycle=False):
    """Print states in "cylc dump" format to STDOUT.

    states = {
        "task_id": {
            "name": name,
            "label": point,
            "state": state,
            "spawned": True|False},
        # ...
    }
    """
    lines = []
    for item in states.values():
        if item['spawned'] in [True, "True", "true"]:
            spawned = 'spawned'
        else:
            spawned = 'unspawned'
        if sort_by_cycle:
            values = [item['label'], item['name'], item['state'], spawned]
        else:
            values = [item['name'], item['label'], item['state'], spawned]
        lines.append(', '.join(values))

    lines.sort()
    for line in lines:
        print line



#!/usr/bin/pyro

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Pyro port scan utilities."""

from multiprocessing import cpu_count, Pool
import sys
from time import sleep
import traceback

import Pyro.errors
import Pyro.core

from cylc.cfgspec.globalcfg import GLOBAL_CFG
import cylc.flags
from cylc.network import PYRO_SUITEID_OBJ_NAME, NO_PASSPHRASE
from cylc.network.connection_validator import ConnValidator, SCAN_HASH
from cylc.network.suite_state import SuiteStillInitialisingError
from cylc.owner import USER
from cylc.registration import RegistrationDB
from cylc.suite_host import get_hostname, is_remote_host


def get_proxy(host, port, pyro_timeout):
    """Return Pyro URL of proxy."""
    proxy = Pyro.core.getProxyForURI(
        'PYROLOC://%s:%s/%s' % (host, port, PYRO_SUITEID_OBJ_NAME))
    proxy._setTimeout(pyro_timeout)
    return proxy


def scan(host=None, db=None, pyro_timeout=None):
    """Scan ports, return a list of suites found: [(port, suite.identify())].

    Note that we could easily scan for a given suite+owner and return its
    port instead of reading port files, but this may not always be fast enough.
    """
    if host is None:
        host = get_hostname()
    base_port = GLOBAL_CFG.get(['pyro', 'base port'])
    last_port = base_port + GLOBAL_CFG.get(['pyro', 'maximum number of ports'])
    if pyro_timeout:
        pyro_timeout = float(pyro_timeout)
    else:
        pyro_timeout = None

    reg_db = RegistrationDB(db)
    results = []
    for port in range(base_port, last_port):
        try:
            proxy = get_proxy(host, port, pyro_timeout)
            conn_val = ConnValidator()
            conn_val.set_default_hash(SCAN_HASH)
            proxy._setNewConnectionValidator(conn_val)
            proxy._setIdentification((USER, NO_PASSPHRASE))
            result = (port, proxy.identify())
        except Pyro.errors.ConnectionDeniedError as exc:
            if cylc.flags.debug:
                print '%s:%s (connection denied)' % (host, port)
            # Back-compat <= 6.4.1
            msg = '  Old daemon at %s:%s?' % (host, port)
            for pphrase in reg_db.load_all_passphrases():
                try:
                    proxy = get_proxy(host, port, pyro_timeout)
                    proxy._setIdentification(pphrase)
                    info = proxy.id()
                    result = (port, {'name': info[0], 'owner': info[1]})
                except Pyro.errors.ConnectionDeniedError:
                    connected = False
                else:
                    connected = True
                    break
            if not connected:
                if cylc.flags.verbose:
                    print >> sys.stderr, msg, "- connection denied (%s)" % exc
                continue
            else:
                if cylc.flags.verbose:
                    print >> sys.stderr, msg, "- connected with passphrase"
        except Pyro.errors.TimeoutError as exc:
            # E.g. Ctrl-Z suspended suite - holds up port scanning!
            if cylc.flags.debug:
                print '%s:%s (connection timed out)' % (host, port)
            print >> sys.stderr, (
                'suite? owner?@%s:%s - connection timed out (%s)' % (
                    host, port, exc))
            continue
        except (Pyro.errors.ProtocolError, Pyro.errors.NamingError) as exc:
            # No suite at this port.
            if cylc.flags.debug:
                print str(exc)
                print '%s:%s (no suite)' % (host, port)
            continue
        except SuiteStillInitialisingError:
            continue
        except Exception as exc:
            if cylc.flags.debug:
                traceback.print_exc()
            raise
        else:
            owner = result[1].get('owner')
            name = result[1].get('name')
            states = result[1].get('states', None)
            if cylc.flags.debug:
                print '   suite:', name, owner
            if states is None:
                # This suite keeps its state info private.
                # Try again with the passphrase if I have it.
                pphrase = reg_db.load_passphrase(name, owner, host)
                if pphrase:
                    try:
                        proxy = get_proxy(host, port, pyro_timeout)
                        conn_val = ConnValidator()
                        conn_val.set_default_hash(SCAN_HASH)
                        proxy._setNewConnectionValidator(conn_val)
                        proxy._setIdentification((USER, pphrase))
                        result = (port, proxy.identify())
                    except Exception:
                        # Nope (private suite, wrong passphrase).
                        if cylc.flags.debug:
                            print '    (wrong passphrase)'
                    else:
                        reg_db.cache_passphrase(
                            name, owner, host, pphrase)
                        if cylc.flags.debug:
                            print '    (got states with passphrase)'
        results.append(result)
    return results


def scan_all(hosts=None, reg_db_path=None, pyro_timeout=None):
    """Scan all hosts."""
    if not hosts:
        hosts = GLOBAL_CFG.get(["suite host scanning", "hosts"])
    # Ensure that it does "localhost" only once
    hosts = set(hosts)
    for host in list(hosts):
        if not is_remote_host(host):
            hosts.remove(host)
            hosts.add("localhost")
    proc_pool_size = GLOBAL_CFG.get(["process pool size"])
    if proc_pool_size is None:
        proc_pool_size = cpu_count()
    if proc_pool_size > len(hosts):
        proc_pool_size = len(hosts)
    proc_pool = Pool(proc_pool_size)
    async_results = {}
    for host in hosts:
        async_results[host] = proc_pool.apply_async(
            scan, [host, reg_db_path, pyro_timeout])
    proc_pool.close()
    scan_results = []
    scan_results_hosts = []
    while async_results:
        sleep(0.05)
        for host, async_result in async_results.items():
            if async_result.ready():
                async_results.pop(host)
                try:
                    res = async_result.get()
                except Exception:
                    if cylc.flags.debug:
                        traceback.print_exc()
                else:
                    scan_results.extend(res)
                    scan_results_hosts.extend([host] * len(res))
    proc_pool.join()
    return zip(scan_results_hosts, scan_results)



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Provide a class to represent a task proxy in a running suite."""

from collections import namedtuple
from copy import copy
from logging import (
    getLevelName, getLogger, CRITICAL, ERROR, WARNING, INFO, DEBUG)
import os
from pipes import quote
from random import randrange
import re
from shutil import rmtree
import time
import traceback

from isodatetime.timezone import get_local_time_zone

from cylc.mkdir_p import mkdir_p
from cylc.cfgspec.globalcfg import GLOBAL_CFG
import cylc.cycling.iso8601
from cylc.envvar import expandvars
import cylc.flags as flags
from cylc.wallclock import (
    get_current_time_string,
    get_seconds_as_interval_string,
    get_time_string_from_unix_time,
    get_unix_time_from_time_string,
    RE_DATE_TIME_FORMAT_EXTENDED,
)
from cylc.host_select import get_task_host
from cylc.job_file import JobFile
from cylc.job_host import RemoteJobHostManager
from cylc.batch_sys_manager import BATCH_SYS_MANAGER
from cylc.owner import is_remote_user, USER
from cylc.poll_timer import PollTimer
from cylc.suite_host import is_remote_host, get_suite_host
from parsec.OrderedDict import OrderedDictWithDefaults
from cylc.mp_pool import SuiteProcPool, SuiteProcContext
from cylc.rundb import CylcSuiteDAO
from cylc.task_id import TaskID
from cylc.task_message import TaskMessage
from parsec.util import pdeepcopy, poverride
from parsec.config import ItemNotFoundError
from cylc.task_state import (
    TaskState, TASK_STATUSES_ACTIVE, TASK_STATUS_WAITING,
    TASK_STATUS_READY, TASK_STATUS_SUBMITTED, TASK_STATUS_SUBMIT_FAILED,
    TASK_STATUS_RUNNING, TASK_STATUS_SUCCEEDED, TASK_STATUS_FAILED)
from cylc.task_outputs import (
    TASK_OUTPUT_STARTED, TASK_OUTPUT_SUCCEEDED, TASK_OUTPUT_FAILED)
from cylc.suite_logging import LOG


CustomTaskEventHandlerContext = namedtuple(
    "CustomTaskEventHandlerContext",
    ["key", "ctx_type", "cmd"])


TaskEventMailContext = namedtuple(
    "TaskEventMailContext",
    ["key", "ctx_type", "event", "mail_from", "mail_to", "mail_smtp"])


TaskJobLogsRegisterContext = namedtuple(
    "TaskJobLogsRegisterContext",
    ["key", "ctx_type"])


TaskJobLogsRetrieveContext = namedtuple(
    "TaskJobLogsRetrieveContext",
    ["key", "ctx_type", "user_at_host", "max_size"])


class TryState(object):
    """Represent the current state of a (re)try."""

    # Memory optimization - constrain possible attributes to this list.
    __slots__ = ["ctx", "delays", "num", "delay", "timeout", "is_waiting"]

    def __init__(self, ctx=None, delays=None):
        self.ctx = ctx
        if delays:
            self.delays = list(delays)
        else:
            self.delays = [0]
        self.num = 0
        self.delay = None
        self.timeout = None
        self.is_waiting = False

    def delay_as_seconds(self):
        """Return the delay as PTnS, where n is number of seconds."""
        return get_seconds_as_interval_string(self.delay)

    def is_delay_done(self, now=None):
        """Is timeout done?"""
        if self.timeout is None:
            return False
        if now is None:
            now = time.time()
        return now > self.timeout

    def is_timeout_set(self):
        """Return True if timeout is set."""
        return self.timeout is not None

    def next(self):
        """Return the next retry delay if there is one, or None otherwise."""
        try:
            self.delay = self.delays[self.num]
        except IndexError:
            return None
        else:
            self.timeout = time.time() + self.delay
            self.num += 1
            return self.delay

    def set_waiting(self):
        """Set waiting flag, while waiting for action to complete."""
        self.delay = None
        self.is_waiting = True
        self.timeout = None

    def unset_waiting(self):
        """Unset waiting flag after an action has completed."""
        self.is_waiting = False

    def timeout_as_str(self):
        """Return the timeout as an ISO8601 date-time string."""
        return get_time_string_from_unix_time(self.timeout)


class TaskProxySequenceBoundsError(ValueError):
    """Error on TaskProxy.__init__ with out of sequence bounds start point."""

    def __str__(self):
        return "Not loading %s (out of sequence bounds)" % self.args[0]


class TaskProxy(object):
    """The task proxy."""

    # RETRY LOGIC:
    #  1) ABSOLUTE SUBMIT NUMBER increments every time a task is
    #  submitted, manually or automatically by (submission or execution)
    # retries; whether or not the task actually begins executing, and is
    # appended to the task log root filename.
    #  2) SUBMISSION TRY NUMBER increments when task job submission
    # fails, if submission retries are configured, but resets to 1 if
    # the task begins executing; and is used for accounting purposes.
    #  3) EXECUTION TRY NUMBER increments only when task execution fails,
    # if execution retries are configured; and is passed to task
    # environments to allow changed behaviour after previous failures.

    # Format string for single line output
    JOB_LOG_FMT_1 = "%(timestamp)s [%(cmd_key)s %(attr)s] %(mesg)s"
    # Format string for multi-line output
    JOB_LOG_FMT_M = "%(timestamp)s [%(cmd_key)s %(attr)s]\n\n%(mesg)s\n"

    CUSTOM_EVENT_HANDLER = "event-handler"
    EVENT_MAIL = "event-mail"
    HEAD_MODE_LOCAL = "local"
    HEAD_MODE_REMOTE = "remote"
    JOB_FILE_BASE = BATCH_SYS_MANAGER.JOB_FILE_BASE
    JOB_KILL = "job-kill"
    JOB_LOGS_RETRIEVE = "job-logs-retrieve"
    JOB_POLL = "job-poll"
    JOB_SUBMIT = "job-submit"
    MANAGE_JOB_LOGS_TRY_DELAYS = (0, 30, 180)  # PT0S, PT30S, PT3M
    NN = "NN"

    LOGGING_LVL_OF = {
        "INFO": INFO,
        "NORMAL": INFO,
        "WARNING": WARNING,
        "ERROR": ERROR,
        "CRITICAL": CRITICAL,
        "DEBUG": DEBUG,
    }
    RE_MESSAGE_TIME = re.compile(
        '\A(.+) at (' + RE_DATE_TIME_FORMAT_EXTENDED + ')\Z')

    TABLE_TASK_JOBS = CylcSuiteDAO.TABLE_TASK_JOBS
    TABLE_TASK_EVENTS = CylcSuiteDAO.TABLE_TASK_EVENTS
    TABLE_TASK_STATES = CylcSuiteDAO.TABLE_TASK_STATES

    POLLED_INDICATOR = "(polled)"

    event_handler_env = {}
    stop_sim_mode_job_submission = False

    def __init__(
            self, tdef, start_point, status=TASK_STATUS_WAITING,
            has_spawned=False, stop_point=None, is_startup=False,
            validate_mode=False, submit_num=0, is_reload_or_restart=False,
            pre_reload_inst=None, message_queue=None):
        self.tdef = tdef
        if submit_num is None:
            self.submit_num = 0
        else:
            self.submit_num = submit_num
        self.validate_mode = validate_mode
        self.message_queue = message_queue

        if is_startup:
            # adjust up to the first on-sequence cycle point
            adjusted = []
            for seq in self.tdef.sequences:
                adj = seq.get_first_point(start_point)
                if adj:
                    # may be None if out of sequence bounds
                    adjusted.append(adj)
            if not adjusted:
                # This task is out of sequence bounds
                raise TaskProxySequenceBoundsError(self.tdef.name)
            self.point = min(adjusted)
            self.cleanup_cutoff = self.tdef.get_cleanup_cutoff_point(
                self.point, self.tdef.intercycle_offsets)
            self.identity = TaskID.get(self.tdef.name, self.point)
        else:
            self.point = start_point
            self.cleanup_cutoff = self.tdef.get_cleanup_cutoff_point(
                self.point, self.tdef.intercycle_offsets)
            self.identity = TaskID.get(self.tdef.name, self.point)

        self.has_spawned = has_spawned

        self.point_as_seconds = None

        # Manually inserted tasks may have a final cycle point set.
        self.stop_point = stop_point

        self.manual_trigger = False
        self.is_manual_submit = False

        self.summary = {
            'latest_message': "",
            'submitted_time': None,
            'submitted_time_string': None,
            'submit_num': self.submit_num,
            'started_time': None,
            'started_time_string': None,
            'finished_time': None,
            'finished_time_string': None,
            'name': self.tdef.name,
            'description': self.tdef.rtconfig['description'],
            'title': self.tdef.rtconfig['title'],
            'label': str(self.point),
            'logfiles': [],
            'job_hosts': {},
            'execution_time_limit': None,
        }
        for lfile in self.tdef.rtconfig['extra log files']:
            self.summary['logfiles'].append(expandvars(lfile))

        self.local_job_file_path = None

        self.retries_configured = False

        self.run_try_state = TryState()
        self.sub_try_state = TryState()
        self.event_handler_try_states = {}
        self.execution_time_limit_poll_try_state = None

        self.db_inserts_map = {
            self.TABLE_TASK_JOBS: [],
            self.TABLE_TASK_STATES: [],
            self.TABLE_TASK_EVENTS: [],
        }
        self.db_updates_map = {
            self.TABLE_TASK_JOBS: [],
            self.TABLE_TASK_STATES: [],
        }

        # TODO - should take suite name from config!
        self.suite_name = os.environ['CYLC_SUITE_NAME']

        # In case task owner and host are needed by db_events_insert()
        # for pre-submission events, set their initial values as if
        # local (we can't know the correct host prior to this because
        # dynamic host selection could be used).
        self.task_host = 'localhost'
        self.task_owner = None
        self.user_at_host = self.task_host

        self.job_vacated = False

        self.submission_poll_timer = None
        self.execution_poll_timer = None

        # An initial db state entry is created at task proxy init. On reloading
        # or restarting the suite, the task proxies already have this db entry.
        if (not self.validate_mode and not is_reload_or_restart and
                self.submit_num == 0):
            self.db_inserts_map[self.TABLE_TASK_STATES].append({
                "time_created": get_current_time_string(),
                "time_updated": get_current_time_string(),
                "status": status})

        if not self.validate_mode and self.submit_num > 0:
            self.db_updates_map[self.TABLE_TASK_STATES].append({
                "time_updated": get_current_time_string(),
                "status": status})

        self.event_hooks = None
        self.sim_mode_run_length = None
        self.set_from_rtconfig()
        self.delayed_start_str = None
        self.delayed_start = None
        self.expire_time_str = None
        self.expire_time = None

        self.state = TaskState(status, self.point, self.identity, tdef,
                               self.db_events_insert, self.db_update_status,
                               self.log)

        if tdef.sequential:
            # Adjust clean-up cutoff.
            p_next = None
            adjusted = []
            for seq in tdef.sequences:
                nxt = seq.get_next_point(self.point)
                if nxt:
                    # may be None if beyond the sequence bounds
                    adjusted.append(nxt)
            if adjusted:
                p_next = min(adjusted)
                if (self.cleanup_cutoff is not None and
                        self.cleanup_cutoff < p_next):
                    self.cleanup_cutoff = p_next

        if is_reload_or_restart and pre_reload_inst is not None:
            self.log(INFO, 'reloaded task definition')
            if pre_reload_inst.state.status in TASK_STATUSES_ACTIVE:
                self.log(WARNING, "job is active with pre-reload settings")
            # Retain some state from my pre suite-reload predecessor.
            self.has_spawned = pre_reload_inst.has_spawned
            self.summary = pre_reload_inst.summary
            self.run_try_state = pre_reload_inst.run_try_state
            self.sub_try_state = pre_reload_inst.sub_try_state
            self.submit_num = pre_reload_inst.submit_num
            self.db_inserts_map = pre_reload_inst.db_inserts_map
            self.db_updates_map = pre_reload_inst.db_updates_map
            # Retain status of outputs.
            for msg, oid in pre_reload_inst.state.outputs.completed.items():
                self.state.outputs.completed[msg] = oid
                try:
                    del self.state.outputs.not_completed[msg]
                except KeyError:
                    pass

    def _get_events_conf(self, key, default=None):
        """Return an events setting from suite then global configuration."""
        for getter in [self.event_hooks, GLOBAL_CFG.get()["task events"]]:
            try:
                value = getter.get(key)
                if value is not None:
                    return value
            except (ItemNotFoundError, KeyError):
                pass
        return default

    def _get_host_conf(self, key, default=None):
        """Return a host setting from suite then global configuration."""
        if self.tdef.rtconfig["remote"].get(key) is not None:
            return self.tdef.rtconfig["remote"][key]
        else:
            try:
                return GLOBAL_CFG.get_host_item(
                    key, self.task_host, self.task_owner)
            except ItemNotFoundError:
                pass
        return default

    def log(self, lvl=INFO, msg=""):
        """Log a message of this task proxy."""
        msg = "[%s] -%s" % (self.identity, msg)
        LOG.log(lvl, msg)

    def command_log(self, ctx):
        """Log an activity for a job of this task proxy."""
        ctx_str = str(ctx)
        if not ctx_str:
            return
        submit_num = self.NN
        if isinstance(ctx.cmd_key, tuple):  # An event handler
            submit_num = ctx.cmd_key[-1]
        job_activity_log = self.get_job_log_path(
            self.HEAD_MODE_LOCAL, submit_num, "job-activity.log")
        try:
            with open(job_activity_log, "ab") as handle:
                handle.write(ctx_str + '\n')
        except IOError as exc:
            LOG.warning(
                "%s: write failed\n%s" % (job_activity_log, exc))
        if ctx.cmd and ctx.ret_code:
            LOG.error(ctx_str)
        elif ctx.cmd:
            LOG.debug(ctx_str)

    def db_events_insert(self, event="", message=""):
        """Record an event to the DB."""
        self.db_inserts_map[self.TABLE_TASK_EVENTS].append({
            "time": get_current_time_string(),
            "event": event,
            "message": message})

    def db_update_status(self):
        """Update suite runtime DB task states table."""
        self.db_updates_map[self.TABLE_TASK_STATES].append({
            "time_updated": get_current_time_string(),
            "submit_num": self.submit_num,
            "try_num": self.run_try_state.num + 1,
            "status": self.state.status})

    def retry_delay_done(self):
        """Is retry delay done? Can I retry now?"""
        now = time.time()
        return (self.run_try_state.is_delay_done(now) or
                self.sub_try_state.is_delay_done(now))

    def ready_to_run(self):
        """Am I in a pre-run state but ready to run?

        Queued tasks are not counted as they've already been deemed ready.

        """
        ready = self.state.is_ready_to_run(self.retry_delay_done(),
                                           self.start_time_reached())
        if ready and self._has_expired():
            self.log(WARNING, 'Task expired (skipping job).')
            self.setup_event_handlers(
                "expired", 'Task expired (skipping job).')
            self.state.set_expired()
            return False
        return ready

    def get_point_as_seconds(self):
        """Compute and store my cycle point as seconds."""
        if self.point_as_seconds is None:
            iso_timepoint = cylc.cycling.iso8601.point_parse(str(self.point))
            self.point_as_seconds = int(iso_timepoint.get(
                "seconds_since_unix_epoch"))
            if iso_timepoint.time_zone.unknown:
                utc_offset_hours, utc_offset_minutes = (
                    get_local_time_zone())
                utc_offset_in_seconds = (
                    3600 * utc_offset_hours + 60 * utc_offset_minutes)
                self.point_as_seconds += utc_offset_in_seconds
        return self.point_as_seconds

    @staticmethod
    def get_offset_as_seconds(offset):
        """Return an ISO interval as seconds."""
        iso_offset = cylc.cycling.iso8601.interval_parse(str(offset))
        return int(iso_offset.get_seconds())

    def start_time_reached(self):
        """Has this task reached its clock trigger time?"""
        if self.tdef.clocktrigger_offset is None:
            return True
        if self.delayed_start is None:
            self.delayed_start = (
                self.get_point_as_seconds() +
                self.get_offset_as_seconds(self.tdef.clocktrigger_offset))
            self.delayed_start_str = get_time_string_from_unix_time(
                self.delayed_start)
        return time.time() > self.delayed_start

    def _has_expired(self):
        """Is this task past its use-by date?"""
        if self.tdef.expiration_offset is None:
            return False
        if self.expire_time is None:
            self.expire_time = (
                self.get_point_as_seconds() +
                self.get_offset_as_seconds(self.tdef.expiration_offset))
            self.expire_time_str = get_time_string_from_unix_time(
                self.expire_time)
        return time.time() > self.expire_time

    def job_submission_callback(self, result):
        """Callback on job submission."""
        if result.out is not None:
            out = ""
            for line in result.out.splitlines(True):
                if line.startswith(
                        BATCH_SYS_MANAGER.CYLC_BATCH_SYS_JOB_ID + "="):
                    self.summary['submit_method_id'] = line.strip().replace(
                        BATCH_SYS_MANAGER.CYLC_BATCH_SYS_JOB_ID + "=", "")
                else:
                    out += line
            result.out = out
        self.command_log(result)

        if result.ret_code == SuiteProcPool.JOB_SKIPPED_FLAG:
            return

        if self.summary['submit_method_id'] and result.ret_code == 0:
            self.job_submission_succeeded()
        else:
            self.job_submission_failed()

    def job_poll_callback(self, cmd_ctx, line):
        """Callback on job poll."""
        ctx = SuiteProcContext(self.JOB_POLL, None)
        ctx.out = line
        ctx.ret_code = 0

        items = line.split("|")
        # See cylc.batch_sys_manager.JobPollContext
        try:
            (
                batch_sys_exit_polled, run_status, run_signal,
                time_submit_exit, time_run, time_run_exit
            ) = items[4:10]
        except IndexError:
            self.summary['latest_message'] = 'poll failed'
            flags.iflag = True
            ctx.cmd = cmd_ctx.cmd  # print original command on failure
            return
        finally:
            self.command_log(ctx)
        if run_status == "1" and run_signal in ["ERR", "EXIT"]:
            # Failed normally
            self.process_incoming_message(
                INFO, TASK_OUTPUT_FAILED, time_run_exit)
        elif run_status == "1" and batch_sys_exit_polled == "1":
            # Failed by a signal, and no longer in batch system
            self.process_incoming_message(
                INFO, TASK_OUTPUT_FAILED, time_run_exit)
            self.process_incoming_message(
                INFO, TaskMessage.FAIL_MESSAGE_PREFIX + run_signal,
                time_run_exit)
        elif run_status == "1":
            # The job has terminated, but is still managed by batch system.
            # Some batch system may restart a job in this state, so don't
            # mark as failed yet.
            self.process_incoming_message(INFO, TASK_OUTPUT_STARTED, time_run)
        elif run_status == "0":
            # The job succeeded
            self.process_incoming_message(
                INFO, TASK_OUTPUT_SUCCEEDED, time_run_exit)
        elif time_run and batch_sys_exit_polled == "1":
            # The job has terminated without executing the error trap
            self.process_incoming_message(
                INFO, TASK_OUTPUT_FAILED, "")
        elif time_run:
            # The job has started, and is still managed by batch system
            self.process_incoming_message(INFO, TASK_OUTPUT_STARTED, time_run)
        elif batch_sys_exit_polled == "1":
            # The job never ran, and no longer in batch system
            self.process_incoming_message(
                INFO, "submission failed", time_submit_exit)
        else:
            # The job never ran, and is in batch system
            self.process_incoming_message(
                INFO, TASK_STATUS_SUBMITTED, time_submit_exit)

    def job_poll_message_callback(self, cmd_ctx, line):
        """Callback on job poll message."""
        ctx = SuiteProcContext(self.JOB_POLL, None)
        ctx.out = line
        try:
            event_time, priority, message = line.split("|")[2:5]
        except ValueError:
            ctx.ret_code = 1
            ctx.cmd = cmd_ctx.cmd  # print original command on failure
        else:
            ctx.ret_code = 0
            self.process_incoming_message(priority, message, event_time)
        self.command_log(ctx)

    def job_kill_callback(self, cmd_ctx, line):
        """Callback on job kill."""
        ctx = SuiteProcContext(self.JOB_KILL, None)
        ctx.out = line
        try:
            ctx.timestamp, _, ctx.ret_code = line.split("|", 2)
        except ValueError:
            ctx.ret_code = 1
            ctx.cmd = cmd_ctx.cmd  # print original command on failure
        else:
            ctx.ret_code = int(ctx.ret_code)
            if ctx.ret_code:
                ctx.cmd = cmd_ctx.cmd  # print original command on failure
        self.command_log(ctx)
        log_lvl = INFO
        log_msg = 'killed'
        if ctx.ret_code:  # non-zero exit status
            log_lvl = WARNING
            log_msg = 'kill failed'
            self.state.kill_failed = True
        elif self.state.status == TASK_STATUS_SUBMITTED:
            self.job_submission_failed()
            flags.iflag = True
        elif self.state.status == TASK_STATUS_RUNNING:
            self.job_execution_failed()
            flags.iflag = True
        else:
            log_lvl = WARNING
            log_msg = (
                'ignoring job kill result, unexpected task state: %s' %
                self.state.status)
        self.summary['latest_message'] = log_msg
        self.log(log_lvl, "job(%02d) %s" % (self.submit_num, log_msg))

    def job_submit_callback(self, cmd_ctx, line):
        """Callback on job submit."""
        ctx = SuiteProcContext(self.JOB_SUBMIT, None)
        ctx.out = line
        items = line.split("|")
        try:
            ctx.timestamp, _, ctx.ret_code = items[0:3]
        except ValueError:
            ctx.ret_code = 1
            ctx.cmd = cmd_ctx.cmd  # print original command on failure
        else:
            ctx.ret_code = int(ctx.ret_code)
            if ctx.ret_code:
                ctx.cmd = cmd_ctx.cmd  # print original command on failure
        self.command_log(ctx)

        if ctx.ret_code == SuiteProcPool.JOB_SKIPPED_FLAG:
            return

        try:
            self.summary['submit_method_id'] = items[3]
        except IndexError:
            self.summary['submit_method_id'] = None
        if self.summary['submit_method_id'] and ctx.ret_code == 0:
            self.job_submission_succeeded()
        else:
            self.job_submission_failed()

    def job_cmd_out_callback(self, cmd_ctx, line):
        """Callback on job command STDOUT/STDERR."""
        job_activity_log = self.get_job_log_path(
            self.HEAD_MODE_LOCAL, self.NN, "job-activity.log")
        if cmd_ctx.cmd_kwargs.get("host") and cmd_ctx.cmd_kwargs.get("user"):
            user_at_host = "(%(user)s@%(host)s) " % cmd_ctx.cmd_kwargs
        elif cmd_ctx.cmd_kwargs.get("host"):
            user_at_host = "(%(host)s) " % cmd_ctx.cmd_kwargs
        elif cmd_ctx.cmd_kwargs.get("user"):
            user_at_host = "(%(user)s@localhost) " % cmd_ctx.cmd_kwargs
        else:
            user_at_host = ""
        try:
            timestamp, _, content = line.split("|")
        except ValueError:
            pass
        else:
            line = "%s %s" % (timestamp, content)
        try:
            with open(job_activity_log, "ab") as handle:
                if not line.endswith("\n"):
                    line += "\n"
                handle.write(user_at_host + line)
        except IOError as exc:
            self.log(WARNING, "%s: write failed\n%s" % (job_activity_log, exc))

    def setup_event_handlers(
            self, event, message, db_update=True, db_event=None, db_msg=None):
        """Set up event handlers."""
        # extra args for inconsistent use between events, logging, and db
        # updates
        db_event = db_event or event
        if db_update:
            self.db_events_insert(event=db_event, message=db_msg)

        if self.tdef.run_mode != 'live':
            return

        self.setup_job_logs_retrieval(event, message)
        self.setup_event_mail(event, message)
        self.setup_custom_event_handlers(event, message)

    def setup_job_logs_retrieval(self, event, _=None):
        """Set up remote job logs retrieval."""
        # TODO - use string constants for event names.
        key2 = (self.JOB_LOGS_RETRIEVE, self.submit_num)
        if (event not in ['failed', 'retry', 'succeeded'] or
                self.user_at_host in [USER + '@localhost', 'localhost'] or
                not self._get_host_conf("retrieve job logs") or
                key2 in self.event_handler_try_states):
            return
        self.event_handler_try_states[key2] = TryState(
            TaskJobLogsRetrieveContext(
                # key
                self.JOB_LOGS_RETRIEVE,
                # ctx_type
                self.JOB_LOGS_RETRIEVE,
                self.user_at_host,
                # max_size
                self._get_host_conf("retrieve job logs max size"),
            ),
            self._get_host_conf("retrieve job logs retry delays", []))

    def setup_event_mail(self, event, _):
        """Event notification, by email."""
        key1 = (self.EVENT_MAIL, event)
        if ((key1, self.submit_num) in self.event_handler_try_states or
                event not in self._get_events_conf("mail events", [])):
            return

        self.event_handler_try_states[(key1, self.submit_num)] = TryState(
            TaskEventMailContext(
                key1,
                self.EVENT_MAIL,  # ctx_type
                event,
                self._get_events_conf(  # mail_from
                    "mail from",
                    "notifications@" + get_suite_host(),
                ),
                self._get_events_conf("mail to", USER),  # mail_to
                self._get_events_conf("mail smtp"),  # mail_smtp
            ),
            self._get_events_conf("mail retry delays", []))

    def setup_custom_event_handlers(self, event, message, only_list=None):
        """Call custom event handlers."""
        handlers = []
        if self.event_hooks[event + ' handler']:
            handlers = self.event_hooks[event + ' handler']
        elif (self._get_events_conf('handlers', []) and
                event in self._get_events_conf('handler events', [])):
            handlers = self._get_events_conf('handlers', [])
        retry_delays = self._get_events_conf(
            'handler retry delays',
            self._get_host_conf("task event handler retry delays", []))
        for i, handler in enumerate(handlers):
            key1 = ("%s-%02d" % (self.CUSTOM_EVENT_HANDLER, i), event)
            if (key1, self.submit_num) in self.event_handler_try_states or (
                    only_list and i not in only_list):
                continue
            cmd = handler % {
                "event": quote(event),
                "suite": quote(self.suite_name),
                "point": quote(str(self.point)),
                "name": quote(self.tdef.name),
                "submit_num": self.submit_num,
                "id": quote(self.identity),
                "message": quote(message),
            }
            if cmd == handler:
                # Nothing substituted, assume classic interface
                cmd = "%s '%s' '%s' '%s' '%s'" % (
                    handler, event, self.suite_name, self.identity, message)
            self.log(DEBUG, "Queueing %s handler: %s" % (event, cmd))
            self.event_handler_try_states[(key1, self.submit_num)] = TryState(
                CustomTaskEventHandlerContext(
                    key1,
                    self.CUSTOM_EVENT_HANDLER,
                    cmd,
                ),
                retry_delays)

    def custom_event_handler_callback(self, result):
        """Callback when a custom event handler is done."""
        self.command_log(result)
        try:
            if result.ret_code == 0:
                del self.event_handler_try_states[result.cmd_key]
            else:
                self.event_handler_try_states[result.cmd_key].unset_waiting()
        except KeyError:
            pass

    def job_submission_failed(self, event_time=None):
        """Handle job submission failure."""
        self.log(ERROR, 'submission failed')
        if event_time is None:
            event_time = get_current_time_string()
        self.db_updates_map[self.TABLE_TASK_JOBS].append({
            "time_submit_exit": get_current_time_string(),
            "submit_status": 1,
        })
        try:
            del self.summary['submit_method_id']
        except KeyError:
            pass
        if self.sub_try_state.next() is None:
            # No submission retry lined up: definitive failure.
            self.summary['finished_time'] = float(
                get_unix_time_from_time_string(event_time))
            self.summary['finished_time_string'] = event_time
            flags.pflag = True
            # See github #476.
            self.setup_event_handlers(
                'submission failed', 'job submission failed')
            self.state.set_submit_failed()
        else:
            # There is a submission retry lined up.
            timeout_str = self.sub_try_state.timeout_as_str()

            delay_msg = "submit-retrying in %s" % (
                self.sub_try_state.delay_as_seconds())
            msg = "submission failed, %s (after %s)" % (delay_msg, timeout_str)
            self.log(INFO, "job(%02d) " % self.submit_num + msg)
            self.summary['latest_message'] = msg
            self.db_events_insert(
                event="submission failed", message=delay_msg)
            # TODO - is this insert redundant with setup_event_handlers?
            self.db_events_insert(
                event="submission failed",
                message="submit-retrying in " + str(self.sub_try_state.delay))
            self.setup_event_handlers(
                "submission retry", "job submission failed, " + delay_msg)
            self.state.set_submit_retry()

    def job_submission_succeeded(self):
        """Handle job submission succeeded."""
        if self.summary.get('submit_method_id') is not None:
            self.log(
                INFO, 'submit_method_id=' + self.summary['submit_method_id'])
        self.log(INFO, 'submission succeeded')
        now = time.time()
        now_string = get_time_string_from_unix_time(now)
        self.db_updates_map[self.TABLE_TASK_JOBS].append({
            "time_submit_exit": now,
            "submit_status": 0,
            "batch_sys_job_id": self.summary.get('submit_method_id')})

        if self.tdef.run_mode == 'simulation':
            # Simulate job execution at this point.
            if self.__class__.stop_sim_mode_job_submission:
                self.state.set_ready_to_submit()
            else:
                self.summary['started_time'] = now
                self.summary['started_time_string'] = now_string
                self.state.set_executing()
            return

        self.summary['started_time'] = None
        self.summary['started_time_string'] = None
        self.summary['finished_time'] = None
        self.summary['finished_time_string'] = None

        self.summary['submitted_time'] = now
        self.summary['submitted_time_string'] = now_string
        self.summary['latest_message'] = TASK_STATUS_SUBMITTED
        self.setup_event_handlers("submitted", 'job submitted',
                                  db_event='submission succeeded')

        if self.state.set_submit_succeeded():
            submit_timeout = self._get_events_conf('submission timeout')
            if submit_timeout:
                self.state.submission_timer_timeout = (
                    self.summary['submitted_time'] + submit_timeout)
            else:
                self.state.submission_timer_timeout = None
            self.submission_poll_timer.set_timer()

    def job_execution_failed(self, event_time=None):
        """Handle a job failure."""
        if event_time is None:
            self.summary['finished_time'] = time.time()
            self.summary['finished_time_string'] = (
                get_time_string_from_unix_time(self.summary['finished_time']))
        else:
            self.summary['finished_time'] = float(
                get_unix_time_from_time_string(event_time))
            self.summary['finished_time_string'] = event_time
        self.db_updates_map[self.TABLE_TASK_JOBS].append({
            "run_status": 1,
            "time_run_exit": self.summary['finished_time_string'],
        })
        self.state.execution_timer_timeout = None
        if self.run_try_state.next() is None:
            # No retry lined up: definitive failure.
            # Note the TASK_STATUS_FAILED output is only added if needed.
            flags.pflag = True
            self.state.set_execution_failed()
            self.setup_event_handlers("failed", 'job failed')
        else:
            # There is a retry lined up
            timeout_str = self.run_try_state.timeout_as_str()
            delay_msg = "retrying in %s" % (
                self.run_try_state.delay_as_seconds())
            msg = "failed, %s (after %s)" % (delay_msg, timeout_str)
            self.log(INFO, "job(%02d) " % self.submit_num + msg)
            self.summary['latest_message'] = msg
            self.setup_event_handlers(
                "retry", "job failed, " + delay_msg, db_msg=delay_msg)
            self.state.set_execution_retry()

    def reset_manual_trigger(self):
        """This is called immediately after manual trigger flag used."""
        if self.manual_trigger:
            self.manual_trigger = False
            self.is_manual_submit = True
            # unset any retry delay timers
            self.run_try_state.timeout = None
            self.sub_try_state.timeout = None

    def set_from_rtconfig(self, cfg=None):
        """Populate task proxy with runtime configuration.

        Some [runtime] config requiring consistency checking on reload,
        and self variables requiring updating for the same.

        """

        if cfg:
            rtconfig = cfg
        else:
            rtconfig = self.tdef.rtconfig

        if not self.retries_configured:
            # configure retry delays before the first try
            self.retries_configured = True
            # TODO - saving the retry delay lists here is not necessary
            # (it can be handled like the polling interval lists).
            if (self.tdef.run_mode == 'live' or
                    (self.tdef.run_mode == 'simulation' and
                        not rtconfig['simulation mode']['disable retries']) or
                    (self.tdef.run_mode == 'dummy' and
                        not rtconfig['dummy mode']['disable retries'])):
                # note that a *copy* of the retry delays list is needed
                # so that all instances of the same task don't pop off
                # the same deque (but copy of rtconfig above solves this).
                self.run_try_state.delays = list(
                    rtconfig['job']['execution retry delays'])
                self.sub_try_state.delays = list(
                    rtconfig['job']['submission retry delays'])

        rrange = rtconfig['simulation mode']['run time range']
        if len(rrange) != 2:
            raise Exception("ERROR, " + self.tdef.name + ": simulation mode " +
                            "run time range should be ISO 8601-compatible")
        try:
            self.sim_mode_run_length = randrange(rrange[0], rrange[1])
        except Exception, exc:
            traceback.print_exc(exc)
            raise Exception(
                "ERROR: simulation mode task run time range must be [MIN,MAX)")

        self.event_hooks = rtconfig['events']

        self.submission_poll_timer = PollTimer(
            copy(rtconfig['job']['submission polling intervals']),
            copy(GLOBAL_CFG.get(['submission polling intervals'])),
            'submission', self.log)

        self.execution_poll_timer = PollTimer(
            copy(rtconfig['job']['execution polling intervals']),
            copy(GLOBAL_CFG.get(['execution polling intervals'])),
            'execution', self.log)

    def submit(self):
        """For "cylc submit". See also "TaskPool.submit_task_jobs"."""

        self.state.set_ready_to_submit()

        # Reset flag so any re-triggering will generate a new job file.
        self.local_job_file_path = None

        cmd_key = self.JOB_SUBMIT
        args = [self.get_job_log_path(
            self.HEAD_MODE_REMOTE, tail=self.JOB_FILE_BASE)]
        stdin_file_paths = [self.get_job_log_path(
            self.HEAD_MODE_LOCAL, tail=self.JOB_FILE_BASE)]

        cmd = ["cylc", cmd_key]
        if cylc.flags.debug:
            cmd.append("--debug")
        remote_mode = False
        for key, value, test_func in [
                ('host', self.task_host, is_remote_host),
                ('user', self.task_owner, is_remote_user)]:
            if test_func(value):
                cmd.append('--%s=%s' % (key, value))
                remote_mode = True
        if remote_mode:
            cmd.append('--remote-mode')
        cmd.append("--")
        cmd += list(args)

        self.log(INFO, "job(%02d) initiate %s" % (self.submit_num, cmd_key))
        ctx = SuiteProcContext(
            cmd_key, cmd, stdin_file_paths=stdin_file_paths)
        return SuiteProcPool.get_inst().put_command(
            ctx, self.job_submission_callback)

    def prep_submit(self, dry_run=False, overrides=None):
        """Prepare job submission.

        Return self on a good preparation.

        """
        if self.tdef.run_mode == 'simulation' or (
                self.local_job_file_path and not dry_run):
            return self

        try:
            job_conf = self._prep_submit_impl(overrides)
            local_job_file_path = self.get_job_log_path(
                self.HEAD_MODE_LOCAL, tail=self.JOB_FILE_BASE)
            JobFile.get_inst().write(local_job_file_path, job_conf)
        except Exception, exc:
            # Could be a bad command template.
            if flags.debug:
                traceback.print_exc()
            self.command_log(SuiteProcContext(
                self.JOB_SUBMIT, '(prepare job file)', err=exc,
                ret_code=1))
            self.job_submission_failed()
            return
        self.local_job_file_path = local_job_file_path

        if dry_run:
            # This will be shown next to submit num in gcylc:
            self.summary['latest_message'] = 'job file written for edit-run'
            self.log(WARNING, self.summary['latest_message'])

        # Return value used by "cylc submit" and "cylc jobscript":
        return self

    def _prep_submit_impl(self, overrides=None):
        """Helper for self.prep_submit."""
        self.log(DEBUG, "incrementing submit number")
        self.submit_num += 1
        self.summary['submit_num'] = self.submit_num
        self.local_job_file_path = None
        self.db_events_insert(event="incrementing submit number")
        self.db_inserts_map[self.TABLE_TASK_JOBS].append({
            "is_manual_submit": self.is_manual_submit,
            "try_num": self.run_try_state.num + 1,
            "time_submit": get_current_time_string(),
        })
        if overrides:
            rtconfig = pdeepcopy(self.tdef.rtconfig)
            poverride(rtconfig, overrides)
        else:
            rtconfig = self.tdef.rtconfig

        self.set_from_rtconfig(rtconfig)

        # construct the job_sub_method here so that a new one is used if
        # the task is re-triggered by the suite operator - so it will
        # get new stdout/stderr logfiles and not overwrite the old ones.

        # dynamic instantiation - don't know job sub method till run time.
        self.summary['batch_sys_name'] = rtconfig['job']['batch system']
        self.summary['execution_time_limit'] = (
            rtconfig['job']['execution time limit'])

        # Determine task host settings now, just before job submission,
        # because dynamic host selection may be used.

        # host may be None (= run task on suite host)
        self.task_host = get_task_host(rtconfig['remote']['host'])
        if not self.task_host:
            self.task_host = 'localhost'
        elif self.task_host != "localhost":
            self.log(INFO, "Task host: " + self.task_host)

        self.task_owner = rtconfig['remote']['owner']

        if self.task_owner:
            self.user_at_host = self.task_owner + "@" + self.task_host
        else:
            self.user_at_host = self.task_host
        self.summary['host'] = self.user_at_host
        self.summary['job_hosts'][self.submit_num] = self.user_at_host
        self.submission_poll_timer.set_host(self.task_host)
        self.execution_poll_timer.set_host(self.task_host)
        try:
            batch_sys_conf = self._get_host_conf('batch systems')[
                rtconfig['job']['batch system']]
        except (TypeError, KeyError):
            batch_sys_conf = OrderedDictWithDefaults()
        if self.summary['execution_time_limit']:
            # Default = 1, 2 and 7 minutes intervals, roughly 1, 3 and 10
            # minutes after time limit exceeded
            self.execution_time_limit_poll_try_state = (
                TryState(delays=batch_sys_conf.get(
                    'execution time limit polling intervals', [60, 120, 420])))

        RemoteJobHostManager.get_inst().init_suite_run_dir(
            self.suite_name, self.user_at_host)
        self.db_updates_map[self.TABLE_TASK_JOBS].append({
            "user_at_host": self.user_at_host,
            "batch_sys_name": self.summary['batch_sys_name'],
        })
        self.is_manual_submit = False

        script, pre_script, post_script = self._get_job_scripts(rtconfig)
        execution_time_limit = rtconfig['job']['execution time limit']
        if execution_time_limit:
            execution_time_limit = float(execution_time_limit)

        # Location of job file, etc
        self._create_job_log_path()

        return {
            'batch system name': rtconfig['job']['batch system'],
            'batch submit command template': (
                rtconfig['job']['batch submit command template']),
            'batch system conf': batch_sys_conf,
            'directives': rtconfig['directives'],
            'execution time limit': execution_time_limit,
            'env-script': rtconfig['env-script'],
            'host': self.task_host,
            'init-script': rtconfig['init-script'],
            'job file path': self.get_job_log_path(
                self.HEAD_MODE_REMOTE, tail=self.JOB_FILE_BASE),
            'job log dir': self.get_job_log_path(),
            'job script shell': rtconfig['job']['shell'],
            'local job file path': self.get_job_log_path(
                self.HEAD_MODE_LOCAL, tail=self.JOB_FILE_BASE),
            'namespace hierarchy': self.tdef.namespace_hierarchy,
            'owner': self.task_owner,
            'post-script': post_script,
            'pre-script': pre_script,
            'remote suite path': (
                rtconfig['remote']['suite definition directory']),
            'runtime environment': rtconfig['environment'],
            'script': script,
            'submit num': self.submit_num,
            'suite name': self.suite_name,
            'task id': self.identity,
            'try number': self.run_try_state.num + 1,
            'work sub-directory': rtconfig['work sub-directory'],
        }

    def _get_job_scripts(self, rtconfig):
        """Return script, pre-script, post-script for a job."""
        script = rtconfig['script']
        pre_script = rtconfig['pre-script']
        post_script = rtconfig['post-script']
        if self.tdef.run_mode == 'dummy':
            # Use dummy script items in dummy mode.
            script = rtconfig['dummy mode']['script']
            if rtconfig['dummy mode']['disable pre-script']:
                pre_script = None
            if rtconfig['dummy mode']['disable post-script']:
                post_script = None
        elif self.tdef.suite_polling_cfg:
            # Automatic suite state polling script
            comstr = "cylc suite-state " + \
                     " --task=" + self.tdef.suite_polling_cfg['task'] + \
                     " --point=" + str(self.point) + \
                     " --status=" + self.tdef.suite_polling_cfg['status']
            for key, fmt in [
                    ('user', ' --%s=%s'),
                    ('host', ' --%s=%s'),
                    ('interval', ' --%s=%d'),
                    ('max-polls', ' --%s=%s'),
                    ('run-dir', ' --%s=%s'),
                    ('template', ' --%s=%s')]:
                if rtconfig['suite state polling'][key]:
                    comstr += fmt % (key, rtconfig['suite state polling'][key])
            comstr += " " + self.tdef.suite_polling_cfg['suite']
            script = "echo " + comstr + "\n" + comstr
        return script, pre_script, post_script

    def _create_job_log_path(self):
        """Create job log directory, etc.

        Create local job directory, and NN symbolic link.
        If NN => 01, remove numbered directories with submit numbers greater
        than 01.
        Return a string in the form "POINT/NAME/SUBMIT_NUM".

        """
        job_file_dir = self.get_job_log_path(self.HEAD_MODE_LOCAL)
        task_log_dir = os.path.dirname(job_file_dir)
        if self.submit_num == 1:
            try:
                names = os.listdir(task_log_dir)
            except OSError:
                pass
            else:
                for name in names:
                    if name not in ["01", self.NN]:
                        rmtree(
                            os.path.join(task_log_dir, name),
                            ignore_errors=True)
        else:
            rmtree(job_file_dir, ignore_errors=True)

        mkdir_p(job_file_dir)
        target = os.path.join(task_log_dir, self.NN)
        source = os.path.basename(job_file_dir)
        try:
            prev_source = os.readlink(target)
        except OSError:
            prev_source = None
        if prev_source == source:
            return
        try:
            if prev_source:
                os.unlink(target)
            os.symlink(source, target)
        except OSError as exc:
            if not exc.filename:
                exc.filename = target
            raise exc

    def handle_submission_timeout(self):
        """Handle submission timeout, only called if TASK_STATUS_SUBMITTED."""
        # Extend timeout so the job can be polled again at next timeout
        # just in case the job is still stuck in a queue
        old_timeout = self.state.submission_timer_timeout
        self.state.submission_timer_timeout = None
        msg = 'job submitted %s ago, but has not started' % (
            get_seconds_as_interval_string(
                old_timeout - self.summary['submitted_time']))
        self.log(WARNING, msg)
        self.setup_event_handlers('submission timeout', msg)
        return True

    def handle_execution_timeout(self):
        """Handle execution timeout, only called if TASK_STATUS_RUNNING."""
        old_timeout = self.state.execution_timer_timeout
        if self.summary['execution_time_limit']:
            try_state = self.execution_time_limit_poll_try_state
            if not try_state.is_timeout_set():
                try_state.next()
            if not try_state.is_delay_done():
                # Don't poll
                return False
            if self.execution_time_limit_poll_try_state.next() is not None:
                # Poll now, and more retries lined up
                return True
            # No more retry lined up, issue execution timeout event
            self.state.execution_timer_timeout = None
        else:
            self.state.execution_timer_timeout = None
        msg = 'job started %s ago, but has not finished' % (
            get_seconds_as_interval_string(
                old_timeout - self.summary['started_time']))
        self.log(WARNING, msg)
        self.setup_event_handlers('execution timeout', msg)
        return True

    def sim_time_check(self):
        """Check simulation time."""
        timeout = self.summary['started_time'] + self.sim_mode_run_length
        if time.time() > timeout:
            if self.tdef.rtconfig['simulation mode']['simulate failure']:
                self.message_queue.put(
                    self.identity, 'NORMAL', TASK_STATUS_SUBMITTED)
                self.message_queue.put(
                    self.identity, 'CRITICAL', TASK_STATUS_FAILED)
            else:
                self.message_queue.put(
                    self.identity, 'NORMAL', TASK_STATUS_SUBMITTED)
                self.message_queue.put(
                    self.identity, 'NORMAL', TASK_STATUS_SUCCEEDED)
            return True
        else:
            return False

    def reject_if_failed(self, message):
        """Reject a message if in the failed state.

        Handle 'enable resurrection' mode.

        """
        if self.state.status == TASK_STATUS_FAILED:
            if self.tdef.rtconfig['enable resurrection']:
                self.log(
                    WARNING,
                    'message receive while failed:' +
                    ' I am returning from the dead!'
                )
                return False
            else:
                self.log(
                    WARNING,
                    'rejecting a message received while in the failed state:'
                )
                self.log(WARNING, '  ' + message)
            return True
        else:
            return False

    def process_incoming_message(
            self, priority, message, polled_event_time=None):
        """Parse an incoming task message and update task state.

        Incoming is e.g. "succeeded at <TIME>".

        Correctly handle late (out of order) message which would otherwise set
        the state backward in the natural order of events.

        """
        is_polled = polled_event_time is not None

        # Log incoming messages with '>' to distinguish non-message log entries
        log_message = '(current:%s)> %s' % (self.state.status, message)
        if polled_event_time is not None:
            log_message += ' %s' % self.POLLED_INDICATOR
        self.log(self.LOGGING_LVL_OF.get(priority, INFO), log_message)

        # Strip the "at TIME" suffix.
        event_time = polled_event_time
        if not event_time:
            match = self.RE_MESSAGE_TIME.match(message)
            if match:
                message, event_time = match.groups()
        if not event_time:
            event_time = get_current_time_string()

        # always update the suite state summary for latest message
        self.summary['latest_message'] = message
        if is_polled:
            self.summary['latest_message'] += " %s" % self.POLLED_INDICATOR
        flags.iflag = True

        if self.reject_if_failed(message):
            # Failed tasks do not send messages unless declared resurrectable
            return

        # Check registered outputs.
        self.state.record_output(message, is_polled)

        if is_polled and self.state.status not in TASK_STATUSES_ACTIVE:
            # A poll result can come in after a task finishes.
            self.log(WARNING, "Ignoring late poll result: task is not active")
            return

        if priority == TaskMessage.WARNING:
            self.setup_event_handlers('warning', message, db_update=False)

        if (message == TASK_OUTPUT_STARTED and
                self.state.status in [TASK_STATUS_READY, TASK_STATUS_SUBMITTED,
                                      TASK_STATUS_SUBMIT_FAILED]):
            if self.job_vacated:
                self.job_vacated = False
                self.log(WARNING, "Vacated job restarted: " + message)
            # Received a 'task started' message
            flags.pflag = True
            self.state.set_executing()
            self.summary['started_time'] = float(
                get_unix_time_from_time_string(event_time))
            self.summary['started_time_string'] = event_time
            self.db_updates_map[self.TABLE_TASK_JOBS].append({
                "time_run": self.summary['started_time_string']})
            if self.summary['execution_time_limit']:
                execution_timeout = self.summary['execution_time_limit']
            else:
                execution_timeout = self._get_events_conf('execution timeout')
            if execution_timeout:
                self.state.execution_timer_timeout = (
                    self.summary['started_time'] + execution_timeout)
            else:
                self.state.execution_timer_timeout = None

            # submission was successful so reset submission try number
            self.sub_try_state.num = 0
            self.setup_event_handlers('started', 'job started')
            self.execution_poll_timer.set_timer()

        elif (message == TASK_OUTPUT_SUCCEEDED and
                self.state.status in [
                    TASK_STATUS_READY, TASK_STATUS_SUBMITTED,
                    TASK_STATUS_SUBMIT_FAILED, TASK_STATUS_RUNNING,
                    TASK_STATUS_FAILED]):
            # Received a 'task succeeded' message
            self.state.execution_timer_timeout = None
            self.state.hold_on_retry = False
            flags.pflag = True
            self.summary['finished_time'] = float(
                get_unix_time_from_time_string(event_time))
            self.summary['finished_time_string'] = event_time
            self.db_updates_map[self.TABLE_TASK_JOBS].append({
                "run_status": 0,
                "time_run_exit": self.summary['finished_time_string'],
            })
            # Update mean elapsed time only on task succeeded.
            self.tdef.update_mean_total_elapsed_time(
                self.summary['started_time'], self.summary['finished_time'])
            self.setup_event_handlers("succeeded", "job succeeded")
            self.state.set_execution_succeeded(is_polled)

        elif (message == TASK_OUTPUT_FAILED and
                self.state.status in [
                    TASK_STATUS_READY, TASK_STATUS_SUBMITTED,
                    TASK_STATUS_SUBMIT_FAILED, TASK_STATUS_RUNNING]):
            # (submit- states in case of very fast submission and execution).
            self.job_execution_failed(event_time)

        elif message.startswith(TaskMessage.FAIL_MESSAGE_PREFIX):
            # capture and record signals sent to task proxy
            self.db_events_insert(event="signaled", message=message)
            signal = message.replace(TaskMessage.FAIL_MESSAGE_PREFIX, "")
            self.db_updates_map[self.TABLE_TASK_JOBS].append(
                {"run_signal": signal})

        elif message.startswith(TaskMessage.VACATION_MESSAGE_PREFIX):
            flags.pflag = True
            self.state.set_state(TASK_STATUS_SUBMITTED)
            self.db_events_insert(event="vacated", message=message)
            self.state.execution_timer_timeout = None
            self.summary['started_time'] = None
            self.summary['started_time_string'] = None
            self.sub_try_state.num = 0
            self.job_vacated = True

        elif message == "submission failed":
            # This can arrive via a poll.
            self.state.submission_timer_timeout = None
            self.job_submission_failed(event_time)

        else:
            # Unhandled messages. These include:
            #  * general non-output/progress messages
            #  * poll messages that repeat previous results
            # Note that all messages are logged already at the top.
            self.log(DEBUG, '(current: %s) unhandled: %s' % (
                self.state.status, message))
            if priority in [CRITICAL, ERROR, WARNING, INFO, DEBUG]:
                priority = getLevelName(priority)
            self.db_events_insert(
                event=("message %s" % str(priority).lower()), message=message)

    def spawn(self, state):
        """Spawn the successor of this task proxy."""
        self.has_spawned = True
        next_point = self.next_point()
        if next_point:
            return TaskProxy(
                self.tdef, next_point, state, False, self.stop_point,
                message_queue=self.message_queue)
        else:
            # next_point instance is out of the sequence bounds
            return None

    def ready_to_spawn(self):
        """Return True if ready to spawn my next-cycle successor.

        A task proxy is never ready to spawn if:
           * it has spawned already
           * its state is submit-failed (avoid running multiple instances
             of a task with bad job submission config).
        Otherwise a task proxy is ready to spawn if either:
           * self.tdef.spawn ahead is True (results in spawning out to max
             active cycle points), OR
           * its state is >= submitted (allows successive instances
             to run concurrently, but not out of order).
        """
        if (self.has_spawned or self.tdef.is_coldstart or
                self.state.status == TASK_STATUS_SUBMIT_FAILED):
            return False
        else:
            return (self.tdef.spawn_ahead or
                    self.state.is_greater_than(TASK_STATUS_READY))

    def get_state_summary(self):
        """Return a dict containing the state summary of this task proxy."""
        self.summary['state'] = self.state.status
        self.summary['spawned'] = str(self.has_spawned)
        self.summary['mean total elapsed time'] = (
            self.tdef.mean_total_elapsed_time)
        return self.summary

    def next_point(self):
        """Return the next cycle point."""
        p_next = None
        adjusted = []
        for seq in self.tdef.sequences:
            nxt = seq.get_next_point(self.point)
            if nxt:
                # may be None if beyond the sequence bounds
                adjusted.append(nxt)
        if adjusted:
            p_next = min(adjusted)
        return p_next

    def get_job_log_path(self, head_mode=None, submit_num=None, tail=None):
        """Return the job log path."""
        args = [str(self.point), self.tdef.name]
        if submit_num is None:
            submit_num = self.submit_num
        try:
            submit_num = "%02d" % submit_num
        except TypeError:
            pass
        if submit_num:
            args.append(submit_num)
        if head_mode == self.HEAD_MODE_LOCAL:
            args.insert(0, GLOBAL_CFG.get_derived_host_item(
                self.suite_name, "suite job log directory"))
        elif head_mode == self.HEAD_MODE_REMOTE:
            args.insert(0, GLOBAL_CFG.get_derived_host_item(
                self.suite_name, 'suite job log directory',
                self.task_host, self.task_owner))
        if tail:
            args.append(tail)
        return os.path.join(*args)




import string

class basetimeall(object):
    def _create_26_entry(self):
        x = self.typ()
        for index, ch in enumerate(string.lowercase):
            x[ch] = index
        assert len(x) == 26
        return x

class timeall(basetimeall):
    def __init__(self, typ):
        self.typ = typ

    def time010_get_keys_from_26_entry(self):
        x = self._create_26_entry()
        y = x.keys()

    def time011_walk_keys(self):
        x = self._create_26_entry()
        for y in x.keys():
            pass

    def time012_walk_reverse_keys(self):
        x = self._create_26_entry()
        for y in reversed(x.keys()):
            pass

    def time013ordereddict_walk_keys_reverse(self):
        x = self._create_26_entry()
        for y in x.keys(reverse=True):
            pass


class timeall1(object):
    def __init__(self, typ):
        self.typ = typ

    def time000_empty(self):
        pass

    def time001_create_empty(self):
        x = self.typ()

    def time001_create_five_entry(self):
        x = self.typ()
        x['a'] = 1
        x['b'] = 2
        x['c'] = 3
        x['d'] = 4
        x['e'] = 5

    def time002_create_26_entry(self):
        x = self.typ()
        for index, ch in enumerate(string.lowercase):
            x[ch] = index
        assert len(x) == 26
        return x

    def time003_create_676_entry(self):
        x = self.typ()
        index = 0
        for ch1 in string.lowercase:
            for ch2 in string.lowercase:
                x[ch1 + ch2] = index
                index += 1
        assert len(x) == 676
        return x


    def _time004_create_17576_entry(self):
        # 17576 items
        x = self.typ()
        index = 0
        for ch1 in string.lowercase:
            for ch2 in string.lowercase:
                for ch3 in string.lowercase:
                    x[ch1 + ch2 + ch3] = index
                    index += 1
        assert len(x) == 17576
        return x

    def time010_get_keys_from_26_entry(self):
        x = self.time002_create_26_entry()
        y = x.keys()

    def time020_pop_5_items_26_entry(self):
        x = self.time002_create_26_entry()
        assert x.pop('f') == 5
        assert x.pop('k') == 10
        assert x.pop('p') == 15
        assert x.pop('u') == 20
        assert x.pop('z') == 25
        assert len(x) == 21

    def time021_pop_26_items_676_entry(self):
        x = self.time003_create_676_entry()
        for k in x.keys():
            if k[1] == 'q':
                i = x.pop(k)
                assert (i % 26) == 16
        assert len(x) == 650

    def time030_popitem_last_26_entry(self):
        x = self.time002_create_26_entry()
        x.popitem()
        assert len(x) == 25

    def time031_popitem_last_676_entry(self):
        x = self.time003_create_676_entry()
        x.popitem()
        assert len(x) == 675

    def time031nodict_popitem_100_676_entry(self):
        x = self.time003_create_676_entry()
        i = x.popitem(100)
        assert i[1] == 100
        assert len(x) == 675

    def time040nodict_walk_26_iteritems(self):
        x = self.time002_create_26_entry()
        index = 0
        for y in x.iteritems():
            assert y[0] == string.lowercase[index]
            assert y[1] == index
            index += 1



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from cylc.cycling.loader import get_interval, get_interval_cls
from cylc.syntax_flags import set_syntax_version, VERSION_PREV, VERSION_NEW
from cylc.task_id import TaskID
import re

# Previous node format.
NODE_PREV_RE = re.compile(
    r"^" +
    r"(" + TaskID.NAME_RE + r")" +
    r"""\s*             # Optional whitespace
        (?:\[           # Begin optional [offset] syntax, start [
         \s*            # Optional whitespace
         T              # T as in T-6, T+1, etc
         \s*            # Optional whitespace
         ([+-])         # Either + or - in e.g. T-6, T+1
         (\s*\w+)       # Offset amount
         \s*            # Optional whitespace
         \]             # End ]
        ){0,1}          # End optional [offset] syntax
        (:[\w-]+){0,1}  # Optional type (e.g. :fail, :finish-all)
        $               # End
    """, re.X)

# Cylc's ISO 8601 format.
NODE_ISO_RE = re.compile(
    r"^" +
    r"(" + TaskID.NAME_RE + r")" +
    r"""(?:\[        # Begin optional [offset] syntax
         (?!T[+-])   # Do not match a 'T-' or 'T+' (this is the old format)
         ([^\]]+)    # Continue until next ']'
         \]          # Stop at next ']'
        )?           # End optional [offset] syntax]
        (:[\w-]+|)$  # Optional type (e.g. :succeed)
     """, re.X)

# Cylc's ISO 8601 initial cycle point based format
NODE_ISO_ICT_RE = re.compile(
    r"^" +
    r"(" + TaskID.NAME_RE + r")" +
    r"""\[           # Begin square bracket syntax
        \^           # Initial cycle point offset marker
        ([^\]]*)     # Optional ^offset syntax
        \]           # End square bracket syntax
        (:[\w-]+|)$  # Optional type (e.g. :succeed)
     """, re.X)

# A potentially non-regular offset, such as foo[01T+P1W].
IRREGULAR_OFFSET_RE = re.compile(
    r"""^            # Start of string
        (            # Begin group
         ..+         # EITHER: Two or more characters
         [+-]P       # Then either +P or -P for start of duration
         .*          # Then anything for the rest of the duration
         |           # OR:
         [^P]+       # No 'P' characters anywhere (e.g. T00).
        )            # End group
        $            # End of string
    """, re.X)


class GraphNodeError(Exception):
    """
    Attributes:
        message - what the problem is.
    """
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


class graphnode(object):
    """A node in the cycle suite.rc dependency graph."""

    # Memory optimization - constrain possible attributes to this list.
    __slots__ = ["offset_is_from_ict", "offset_is_irregular",
                 "is_absolute", "special_output", "output",
                 "name", "intercycle", "offset_string"]

    def __init__(self, node, base_interval=None):
        node_in = node
        # Get task name and properties from a graph node name.

        # Graph node name is task name optionally followed by:
        # - output label: foo:m1
        # - intercycle dependence: foo[T-6]
        # These may be combined: foo[T-6]:m1
        # Task may be defined at initial cycle point: foo[^]
        # or relative to initial cycle point: foo[^+P1D]

        self.offset_is_from_ict = False
        self.offset_is_irregular = False
        self.is_absolute = False

        m = re.match(NODE_ISO_ICT_RE, node)
        if m:
            # node looks like foo[^], foo[^-P4D], foo[^]:fail, etc.
            self.is_absolute = True
            name, offset_string, outp = m.groups()
            self.offset_is_from_ict = True
            sign = ""
            prev_format = False
            # Can't always set syntax here, as we use [^] for backwards comp.
            if offset_string:
                set_syntax_version(
                    VERSION_NEW, "graphnode: %s: ISO 8601 offset" % node)
        else:
            m = re.match(NODE_ISO_RE, node)
            if m:
                # node looks like foo, foo:fail, foo[-PT6H], foo[-P4D]:fail...
                name, offset_string, outp = m.groups()
                sign = ""
                prev_format = False
                if offset_string:
                    set_syntax_version(
                        VERSION_NEW, "graphnode: %s: ISO 8601 offset" % node)
            else:
                m = re.match(NODE_PREV_RE, node)
                if not m:
                    raise GraphNodeError('Illegal graph node: ' + node)
                # node looks like foo[T-6], foo[T-12]:fail...
                name, sign, offset_string, outp = m.groups()
                if sign and offset_string:
                    offset_string = sign + offset_string
                prev_format = True
                # Strip :succeed etc. off here, to avoid double warning for
                # adding edges (with :succeed) and getting nodes (without).
                set_syntax_version(
                    VERSION_PREV,
                    "graphnode %s: old-style offset" % re.sub(':.*', '', node)
                )

        if outp:
            self.special_output = True
            self.output = outp[1:]  # strip ':'
        else:
            self.special_output = False
            self.output = None

        if name:
            self.name = name
        else:
            raise GraphNodeError('Illegal graph node: ' + node)

        if self.offset_is_from_ict and not offset_string:
            offset_string = str(get_interval_cls().get_null_offset())
        if offset_string:
            self.intercycle = True
            if prev_format:
                self.offset_string = str(
                    base_interval.get_inferred_child(offset_string))
            else:
                if IRREGULAR_OFFSET_RE.search(offset_string):
                    self.offset_string = offset_string
                    self.offset_is_irregular = True
                else:
                    self.offset_string = str(
                        (get_interval(offset_string)).standardise())
        else:
            self.intercycle = False
            self.offset_string = None



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Common logic for "cylc run" and "cylc restart" CLI."""

import sys

from cylc.option_parsers import CylcOptionParser as COP
import cylc.flags
from cylc.scheduler import Scheduler

RUN_DOC = r"""cylc [control] run|start [OPTIONS] ARGS

Start a suite run from scratch, wiping out any previous suite state. To
restart from a previous state see 'cylc restart --help'.

The scheduler runs in daemon mode unless you specify --no-detach or --debug.

Any dependence on cycle points earlier than the start cycle point is ignored.

A "cold start" (the default) starts from the suite initial cycle point
(specified in the suite.rc or on the command line).  Any dependence on tasks
prior to the suite initial cycle point is ignored.

A "warm start" (-w/--warm) starts from a given cycle point later than the suite
initial cycle point (specified in the suite.rc).  Any dependence on tasks prior
to the given warm start cycle point is ignored.  The suite initial cycle point
is preserved."""

RESTART_DOC = r"""cylc [control] restart [OPTIONS] ARGS

Start a suite run from the previous state. To start from scratch (cold or warm
start) see the 'cylc run' command.

The scheduler runs in daemon mode unless you specify n/--no-detach or --debug.

Tasks recorded as submitted or running are polled at start-up to determine what
happened to them while the suite was down."""

SUITE_NAME_ARG_DOC = ("REG", "Suite name")
START_POINT_ARG_DOC = (
    "[START_POINT]",
    "Initial cycle point or 'now';\n" +
    " " * 31 +  # 20 + len("START_POINT")
    "overrides the suite definition.")


def main(is_restart=False):
    """CLI main."""
    options, args = parse_commandline(is_restart)
    scheduler = Scheduler(is_restart, options, args)
    scheduler.start()


def parse_commandline(is_restart):
    """Parse CLI for "cylc run" or "cylc restart"."""
    if is_restart:
        parser = COP(RESTART_DOC, jset=True, argdoc=[SUITE_NAME_ARG_DOC])
    else:
        parser = COP(
            RUN_DOC, jset=True,
            argdoc=[SUITE_NAME_ARG_DOC, START_POINT_ARG_DOC])

    parser.add_option(
        "--non-daemon", help="(deprecated: use --no-detach)",
        action="store_true", default=False, dest="no_detach")

    parser.add_option(
        "-n", "--no-detach", help="Do not daemonize the suite",
        action="store_true", default=False, dest="no_detach")

    parser.add_option(
        "-a", "--no-auto-shutdown", help="Do not shut down"
        " the suite automatically when all tasks have finished."
        " This flag overrides the corresponding suite config item.",
        action="store_true", default=False, dest="no_auto_shutdown")

    parser.add_option(
        "--profile", help="Output profiling (performance) information",
        action="store_true", default=False, dest="profile_mode")

    if is_restart:
        parser.add_option(
            "--checkpoint",
            help="Use specified checkpoint to restart",
            metavar="CHECKPOINT", action="store", dest="checkpoint")

        parser.add_option(
            "--ignore-final-cycle-point",
            help=(
                "Ignore the final cycle point in the suite run database. " +
                "If one is specified in the suite definition it will " +
                "be used, however."),
            action="store_true", default=False, dest="ignore_stop_point")

        parser.add_option(
            "--ignore-initial-cycle-point",
            help=(
                "Ignore the initial cycle point in the suite run database. " +
                "If one is specified in the suite definition it will " +
                "be used, however."),
            action="store_true", default=False, dest="ignore_start_point")
    else:
        parser.add_option(
            "-w", "--warm",
            help="Warm start the suite. "
                 "The default is to cold start.",
            action="store_true", default=False, dest="warm")

        parser.add_option(
            "--ict",
            help="Does nothing, option for backward compatibility only",
            action="store_true", default=False, dest="set_ict")

    parser.add_option(
        "--until",
        help=("Shut down after all tasks have PASSED " +
              "this cycle point."),
        metavar="CYCLE_POINT", action="store", dest="final_point_string")

    parser.add_option(
        "--hold",
        help="Hold (don't run tasks) immediately on starting.",
        action="store_true", default=False, dest="start_held")

    parser.add_option(
        "--hold-after",
        help="Hold (don't run tasks) AFTER this cycle point.",
        metavar="CYCLE_POINT", action="store", dest="hold_point_string")

    parser.add_option(
        "-m", "--mode",
        help="Run mode: live, simulation, or dummy; default is live.",
        metavar="STRING", action="store", default='live', dest="run_mode",
        choices=["live", "dummy", "simulation"])

    parser.add_option(
        "--reference-log",
        help="Generate a reference log for use in reference tests.",
        action="store_true", default=False, dest="genref")

    parser.add_option(
        "--reference-test",
        help="Do a test run against a previously generated reference log.",
        action="store_true", default=False, dest="reftest")

    options, args = parser.parse_args()

    if not is_restart and options.warm and len(args) < 2:
        # Warm start must have a start point
        sys.exit(parser.get_usage())

    return options, args



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""cylc support for the Jinja2 template processor

Importing code should catch ImportError in case Jinja2 is not installed.
"""

from glob import glob
import os
import sys
from jinja2 import (
    Environment,
    FileSystemLoader,
    TemplateError,
    UndefinedError,
    StrictUndefined)
import cylc.flags


def jinja2process(flines, dir_, template_vars=None):
    """Pass configure file through Jinja2 processor."""
    env = Environment(
        loader=FileSystemLoader(dir_),
        undefined=StrictUndefined,
        extensions=['jinja2.ext.do'])

    # Load any custom Jinja2 filters in the suite definition directory
    # Example: a filter to pad integer values some fill character:
    # |(file SUITE_DEFINIION_DIRECTORY/Jinja2/foo.py)
    # |  #!/usr/bin/env python
    # |  def foo( value, length, fillchar ):
    # |     return str(value).rjust( int(length), str(fillchar) )
    for fdir in [
            os.path.join(os.environ['CYLC_DIR'], 'lib', 'Jinja2Filters'),
            os.path.join(dir_, 'Jinja2Filters'),
            os.path.join(os.environ['HOME'], '.cylc', 'Jinja2Filters')]:
        if os.path.isdir(fdir):
            sys.path.append(os.path.abspath(fdir))
            for name in glob(os.path.join(fdir, '*.py')):
                fname = os.path.splitext(os.path.basename(name))[0]
                # TODO - EXCEPTION HANDLING FOR LOADING CUSTOM FILTERS
                module = __import__(fname)
                env.filters[fname] = getattr(module, fname)

    # Import SUITE HOST USER ENVIRONMENT into template:
    # (usage e.g.: {{environ['HOME']}}).
    env.globals['environ'] = os.environ

    # load file lines into a template, excluding '#!jinja2' so
    # that '#!cylc-x.y.z' rises to the top.
    # CALLERS SHOULD HANDLE JINJA2 TEMPLATESYNTAXERROR AND TEMPLATEERROR
    # try:
    # except Exception as exc:
    #     # This happens if we use an unknown Jinja2 filter, for example.
    #     # TODO - THIS IS CAUGHT BY VALIDATE BUT NOT BY VIEW COMMAND...
    #     raise TemplateError(exc)
    if cylc.flags.verbose and template_vars:
        print 'Setting Jinja2 template variables:'
        for item in sorted(template_vars.items()):
            print '    + %s=%s' % item

    # CALLERS SHOULD HANDLE JINJA2 TEMPLATESYNTAXERROR AND TEMPLATEERROR
    # AND TYPEERROR (e.g. for not using "|int" filter on number inputs.
    # Convert unicode to plain str, ToDo - still needed for parsec?)

    suiterc = []
    template = env.from_string('\n'.join(flines[1:]))
    for line in str(template.render(template_vars)).splitlines():
        # Jinja2 leaves blank lines where source lines contain
        # only Jinja2 code; this matters if line continuation
        # markers are involved, so we remove blank lines here.
        if not line.strip():
            continue
            # restoring newlines here is only necessary for display by
        # the cylc view command:
        # ##suiterc.append(line + '\n')
        suiterc.append(line)

    return suiterc



#############################################################################
#
#	Pyro global constants
#
#	This is part of "Pyro" - Python Remote Objects
#	which is (c) Irmen de Jong - irmen@razorvine.net
#
#############################################################################


# General Pyro Version String ####
VERSION = '3.16'

# Fixed (internal) GUIDs
INTERNAL_DAEMON_GUID='c0000000'+'01100000'+'10000000'+'10000001'

# Pyro names for the standard Services
NAMESERVER_NAME		= ":Pyro.NameServer"
EVENTSERVER_NAME	= ":Pyro.EventService"

# Pyro traceback attribute for remote exceptions
TRACEBACK_ATTRIBUTE	= "remote_stacktrace"


#### Remote Invocation Flags (bit flags) ####

RIF_Varargs  = (1<<0)		# for '*args' syntax
RIF_Keywords = (1<<1)		# for '**keywords' syntax
RIF_Oneway   = (1<<2)		# for oneway (no result) messages - currently internal use only
RIF_VarargsAndKeywords = RIF_Varargs | RIF_Keywords


#### Reasons why a connection may be denied ####
DENIED_UNSPECIFIED=0
DENIED_SERVERTOOBUSY=1
DENIED_HOSTBLOCKED=2
DENIED_SECURITY=3

deniedReasons={
	DENIED_UNSPECIFIED:'unspecified reason',
	DENIED_SERVERTOOBUSY:'server too busy',
	DENIED_HOSTBLOCKED:'host blocked',
	DENIED_SECURITY:'security reasons' 
	}

# special config items
CFGITEM_PYRO_INITIALIZED = "_PYRO_INITIALIZED"

# NS roles
NSROLE_SINGLE=0
NSROLE_PRIMARY=1
NSROLE_SECONDARY=2



# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#-------------------------------------------------------------------------------

import os
import sqlite3
import subprocess
import sys


def main(argv):

    if len(argv) != 2:
        print >> sys.stderr, "Incorrect number of args"
        sys.exit(1)

    sname = argv[0]
    rundir = argv[1]

    p = subprocess.Popen("cylc cat-state " + sname, shell=True, 
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    state, err = p.communicate()

    if p.returncode > 0:
        print >> sys.stderr, err
        sys.exit(1)

    db = (os.sep).join([rundir, sname, "cylc-suite.db"])
    cnx = sqlite3.Connection(db)
    cur = cnx.cursor()

    state = state.split("\n")
    states_begun = False

    qbase = "select status from task_states where name==? and cycle==?"

    error_states = []

    for line in state:
        if states_begun and line is not '':
            line2 = line.split(':')
            task_and_cycle = line2[0].strip().split(".")
            status = line2[1].split(',')[0].strip().split("=")[1]
            # query db and compare result
            res = []
            try:
                cur.execute(qbase, [task_and_cycle[0], task_and_cycle[1]])
                next = cur.fetchmany()
                while next:
                    res.append(next[0])
                    next = cur.fetchmany()
            except:
                sys.stderr.write("unable to query suite database\n")
                sys.exit(1)
            if not res[0][0] == status:
                error_states.append(line + ": state retrieved " + str(res[0][0]))
        elif line == "Begin task states":
            states_begun = True

    cnx.close()

    if error_states:
        print >> sys.stderr, "The following task states were not consistent with the database:"
        for line in error_states:
            print >> sys.stderr, line
        sys.exit(1)
    else:
        sys.exit(0)


if __name__ == "__main__":
   main(sys.argv[1:])



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os, sys
# parse:
sys.path.append( os.path.join( os.path.dirname(os.path.abspath(__file__)), '..' ))
# cylc:
sys.path.append( os.path.join( os.path.dirname(os.path.abspath(__file__)), '../..' ))

from cfgspec import SPEC
from config import config
import cylc.flags

cylc.flags.verbose = True
class testcfg( config ):
    def check( self, sparse ):
        # TEMPORARY EXAMPLE
        if 'missing item' not in self.sparse.keys():
            print "missing item is MISSING!!!!"

cfg = testcfg( SPEC )
strict = False
cfg.loadcfg( os.path.join( os.path.dirname( __file__ ), 'site.rc' )) # TODO: test strict=False (fail but accept defaults)
cfg.loadcfg( os.path.join( os.path.dirname( __file__ ), 'user.rc' ))

cfg.printcfg()
#print
#cfg.printcfg( ['list values'] )
#print
#cfg.printcfg( ['list values', 'integers'] )
#print
#cfg.printcfg( ['single values', 'strings with internal comments'] )



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
import sys
import cPickle as pickle
import threading

from cylc.broadcast_report import (
    get_broadcast_change_iter,
    get_broadcast_change_report,
    get_broadcast_bad_options_report)
from cylc.cycling.loader import get_point, standardise_point_string
from cylc.wallclock import get_current_time_string
from cylc.network import PYRO_BCAST_OBJ_NAME
from cylc.network.pyro_base import PyroClient, PyroServer
from cylc.network import check_access_priv
from cylc.suite_logging import LOG
from cylc.task_id import TaskID
from cylc.rundb import CylcSuiteDAO


class BroadcastServer(PyroServer):
    """Server-side suite broadcast interface.

    Examples:
    self.settings['*']['root'] = {'environment': {'FOO': 'bar'}}
    self.settings['20100808T06Z']['root'] = {'command scripting': 'stuff'}
    """

    _INSTANCE = None
    ALL_CYCLE_POINTS_STRS = ["*", "all-cycle-points", "all-cycles"]
    REC_SECTION = re.compile(r"\[([^\]]+)\]")
    TABLE_BROADCAST_EVENTS = CylcSuiteDAO.TABLE_BROADCAST_EVENTS
    TABLE_BROADCAST_STATES = CylcSuiteDAO.TABLE_BROADCAST_STATES

    @classmethod
    def get_inst(cls, linearized_ancestors=None):
        """Return a singleton instance.

        On 1st call, instantiate the singleton.
        Argument linearized_ancestors is only relevant on 1st call.

        """
        if cls._INSTANCE is None:
            cls._INSTANCE = cls(linearized_ancestors)
        return cls._INSTANCE

    def __init__(self, linearized_ancestors):
        super(BroadcastServer, self).__init__()
        self.log = LOG
        self.settings = {}
        self.db_inserts_map = {
            self.TABLE_BROADCAST_EVENTS: [],
            self.TABLE_BROADCAST_STATES: []}
        self.db_deletes_map = {
            self.TABLE_BROADCAST_STATES: []}
        self.linearized_ancestors = linearized_ancestors
        self.lock = threading.RLock()

    def _prune(self):
        """Remove empty leaves left by unsetting broadcast values.

        Return a list of pruned settings in the form:

        [
            ["20200202", "foo", "command scripting"],
            ["20020202", "bar", "environment", "BAR"],
        ]
        """
        with self.lock:
            prunes = []
            stuff_stack = [([], self.settings, True)]
            while stuff_stack:
                keys, stuff, is_new = stuff_stack.pop()
                if is_new:
                    stuff_stack.append((keys, stuff, False))
                    for key, value in stuff.items():
                        if isinstance(value, dict):
                            stuff_stack.append((keys + [key], value, True))
                else:
                    for key, value in stuff.items():
                        if value in [None, {}]:
                            del stuff[key]
                            prunes.append(keys + [key])
            return prunes

    def _addict(self, target, source):
        """Recursively add source dict to target dict."""
        for key, val in source.items():
            if isinstance(val, dict):
                if key not in target:
                    target[key] = {}
                self._addict(target[key], val)
            else:
                target[key] = source[key]

    def put(self, point_strings, namespaces, settings):
        """Add new broadcast settings (server side interface).

        Return a tuple (modified_settings, bad_options) where:
          modified_settings is list of modified settings in the form:
            [("20200202", "foo", {"command scripting": "true"}, ...]
          bad_options is as described in the docstring for self.clear().
        """
        check_access_priv(self, 'full-control')
        self.report('broadcast_put')
        modified_settings = []
        bad_point_strings = []
        bad_namespaces = []

        with self.lock:
            for setting in settings:
                for point_string in point_strings:
                    # Standardise the point and check its validity.
                    bad_point = False
                    try:
                        point_string = standardise_point_string(point_string)
                    except Exception as exc:
                        if point_string != '*':
                            bad_point_strings.append(point_string)
                            bad_point = True
                    if not bad_point and point_string not in self.settings:
                        self.settings[point_string] = {}
                    for namespace in namespaces:
                        if namespace not in self.linearized_ancestors:
                            bad_namespaces.append(namespace)
                        elif not bad_point:
                            if namespace not in self.settings[point_string]:
                                self.settings[point_string][namespace] = {}
                            self._addict(
                                self.settings[point_string][namespace],
                                setting)
                            modified_settings.append(
                                (point_string, namespace, setting))

        # Log the broadcast
        self._append_db_queue(modified_settings)
        self.log.info(get_broadcast_change_report(modified_settings))

        bad_options = {}
        if bad_point_strings:
            bad_options["point_strings"] = bad_point_strings
        if bad_namespaces:
            bad_options["namespaces"] = bad_namespaces
        return modified_settings, bad_options

    def get(self, task_id=None):
        """Retrieve all broadcast variables that target a given task ID."""
        check_access_priv(self, 'full-read')
        self.report('broadcast_get')
        if not task_id:
            # all broadcast settings requested
            return self.settings
        name, point_string = TaskID.split(task_id)

        ret = {}
        # The order is:
        #    all:root -> all:FAM -> ... -> all:task
        # -> tag:root -> tag:FAM -> ... -> tag:task
        for cycle in self.ALL_CYCLE_POINTS_STRS + [point_string]:
            if cycle not in self.settings:
                continue
            for namespace in reversed(self.linearized_ancestors[name]):
                if namespace in self.settings[cycle]:
                    self._addict(ret, self.settings[cycle][namespace])
        return ret

    def expire(self, cutoff):
        """Clear all settings targeting cycle points earlier than cutoff."""
        point_strings = []
        cutoff_point = None
        if cutoff is not None:
            cutoff_point = get_point(str(cutoff))
        with self.lock:
            for point_string in self.settings:
                if cutoff_point is None or (
                        point_string not in self.ALL_CYCLE_POINTS_STRS and
                        get_point(point_string) < cutoff_point):
                    point_strings.append(point_string)
        if not point_strings:
            return (None, {"expire": [cutoff]})
        return self.clear(point_strings=point_strings)

    def clear(self, point_strings=None, namespaces=None, cancel_settings=None):
        """Clear settings globally, or for listed namespaces and/or points.

        Return a tuple (modified_settings, bad_options), where:
        * modified_settings is similar to the return value of the "put" method,
          but for removed settings.
        * bad_options is a dict in the form:
              {"point_strings": ["20020202", ..."], ...}
          The dict is only populated if there are options not associated with
          previous broadcasts. The keys can be:
          * point_strings: a list of bad point strings.
          * namespaces: a list of bad namespaces.
          * cancel: a list of tuples. Each tuple contains the keys of a bad
            setting.
        """
        # If cancel_settings defined, only clear specific settings
        cancel_keys_list = self._settings_to_keys_list(cancel_settings)

        # Clear settings
        modified_settings = []
        with self.lock:
            for point_string, point_string_settings in self.settings.items():
                if point_strings and point_string not in point_strings:
                    continue
                for namespace, namespace_settings in (
                        point_string_settings.items()):
                    if namespaces and namespace not in namespaces:
                        continue
                    stuff_stack = [([], namespace_settings)]
                    while stuff_stack:
                        keys, stuff = stuff_stack.pop()
                        for key, value in stuff.items():
                            if isinstance(value, dict):
                                stuff_stack.append((keys + [key], value))
                            elif (not cancel_keys_list or
                                    keys + [key] in cancel_keys_list):
                                stuff[key] = None
                                setting = {key: value}
                                for rkey in reversed(keys):
                                    setting = {rkey: setting}
                                modified_settings.append(
                                    (point_string, namespace, setting))

        # Prune any empty branches
        bad_options = self._get_bad_options(
            self._prune(), point_strings, namespaces, cancel_keys_list)

        # Log the broadcast
        self._append_db_queue(modified_settings, is_cancel=True)
        self.log.info(
            get_broadcast_change_report(modified_settings, is_cancel=True))
        if bad_options:
            self.log.error(get_broadcast_bad_options_report(bad_options))

        return (modified_settings, bad_options)

    @staticmethod
    def _settings_to_keys_list(settings):
        """Return a list containing each setting dict keys as a list.

        E.g. Each setting in settings may look like:
        {"foo": {"bar": {"baz": 1}}}

        An element of the returned list will look like:
        ["foo", "bar", "baz"]

        """
        keys_list = []
        if settings:
            for setting in settings:
                stuff_stack = [([], setting)]
                while stuff_stack:
                    keys, stuff = stuff_stack.pop()
                    for key, value in stuff.items():
                        if isinstance(value, dict):
                            stuff_stack.append((keys + [key], value))
                        else:
                            keys_list.append(keys + [key])
        return keys_list

    def load_state(self, point, namespace, key, value):
        """Load broadcast variables from runtime DB broadcast states row."""
        sections = []
        if "]" in key:
            sections = self.REC_SECTION.findall(key)
            key = key.rsplit(r"]", 1)[-1]
        with self.lock:
            self.settings.setdefault(point, {})
            self.settings[point].setdefault(namespace, {})
            dict_ = self.settings[point][namespace]
            for section in sections:
                dict_.setdefault(section, {})
                dict_ = dict_[section]
            dict_[key] = value

    @classmethod
    def _get_bad_options(
            cls, prunes, point_strings, namespaces, cancel_keys_list):
        """Return unpruned namespaces and/or point_strings options."""
        cancel_keys_list = [
            tuple(cancel_keys) for cancel_keys in cancel_keys_list]
        bad_options = {
            "point_strings": None, "namespaces": None, "cancel": None}
        # 1. Populate the bad_options dict where applicable.
        # 2. Remove keys if they are found in "prunes".
        # 3. Remove key in bad_options if it becomes empty.
        for opt_name, opt_list, opt_test in [
                ("point_strings", point_strings, cls._point_string_in_prunes),
                ("namespaces", namespaces, cls._namespace_in_prunes),
                ("cancel", cancel_keys_list, cls._cancel_keys_in_prunes)]:
            if opt_list:
                bad_options[opt_name] = set(opt_list)
                for opt in opt_list:
                    if opt_test(prunes, opt):
                        bad_options[opt_name].discard(opt)
                        if not bad_options[opt_name]:
                            break
        for key, value in bad_options.items():
            if value:
                bad_options[key] = list(value)
            else:
                del bad_options[key]
        return bad_options

    @staticmethod
    def _point_string_in_prunes(prunes, point_string):
        """Is point_string pruned?"""
        return point_string in [prune[0] for prune in prunes]

    @staticmethod
    def _namespace_in_prunes(prunes, namespace):
        """Is namespace pruned?"""
        return namespace in [prune[1] for prune in prunes if prune[1:]]

    @staticmethod
    def _cancel_keys_in_prunes(prunes, cancel_keys):
        """Is cancel_keys pruned?"""
        return (list(cancel_keys) in
                [prune[2:] for prune in prunes if prune[2:]])

    def _append_db_queue(self, modified_settings, is_cancel=False):
        """Update the queue to the runtime DB."""
        now = get_current_time_string(display_sub_seconds=True)
        for broadcast_change in (
                get_broadcast_change_iter(modified_settings, is_cancel)):
            broadcast_change["time"] = now
            self.db_inserts_map[self.TABLE_BROADCAST_EVENTS].append(
                broadcast_change)
            if is_cancel:
                self.db_deletes_map[self.TABLE_BROADCAST_STATES].append({
                    "point": broadcast_change["point"],
                    "namespace": broadcast_change["namespace"],
                    "key": broadcast_change["key"]})
                # Delete statements are currently executed before insert
                # statements, so we should clear out any insert statements that
                # are deleted here.
                # (Not the most efficient logic here, but unless we have a
                # large number of inserts, then this should not be a big
                # concern.)
                inserts = []
                for insert in self.db_inserts_map[self.TABLE_BROADCAST_STATES]:
                    if any([insert[key] != broadcast_change[key] for key in
                            ["point", "namespace", "key"]]):
                        inserts.append(insert)
                self.db_inserts_map[self.TABLE_BROADCAST_STATES] = inserts
            else:
                self.db_inserts_map[self.TABLE_BROADCAST_STATES].append({
                    "point": broadcast_change["point"],
                    "namespace": broadcast_change["namespace"],
                    "key": broadcast_change["key"],
                    "value": broadcast_change["value"]})


class BroadcastClient(PyroClient):
    """Client-side suite broadcast interface."""

    target_server_object = PYRO_BCAST_OBJ_NAME

    def broadcast(self, cmd, *args):
        return self.call_server_func(cmd, *args)




import sys
import string
from _ordereddict import ordereddict

def do_empty():
    d = ordereddict()

def do_small():
    d = ordereddict()
    for i in range(5):
        d[i]  = i

def do_one():
    d = ordereddict()
    index = 0
    for ch1 in string.lowercase:
        for ch2 in string.lowercase:
            for ch3 in string.lowercase:
                d[ch1 + ch2] = index
                index += 1

for i in range(1000000):
    if i % 100 == 0:
        print '\r %8d' % ( i ),
        sys.stdout.flush()
    do_one()



#!/usr/bin/env python

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
# Copyright (C) 2008-2016 NIWA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Global syntax flags used in cylc"""

import sys

VERSION_PREV = "pre-cylc-6"  # < cylc-6
VERSION_NEW = "post-cylc-6"  # cylc 6 +

DEPRECATED_SYNTAX_WARNING = {
    VERSION_PREV: "WARNING: pre cylc 6 syntax is deprecated: {0}\n"
}


class SyntaxVersion(object):

    """Store the syntax version used in the suite.rc."""
    VERSION_REASON = None
    VERSION = None
    WARNING_MESSAGES = set()


class SyntaxVersionError(ValueError):

    """Raise an error if conflicting syntax versions are used."""

    ERROR_CONFLICTING_SYNTAX_VERSIONS = (
        "Conflicting syntax: %s syntax (%s) vs %s syntax (%s)")

    def __str__(self):
        return (self.ERROR_CONFLICTING_SYNTAX_VERSIONS % (
                SyntaxVersion.VERSION, SyntaxVersion.VERSION_REASON,
                self.args[0], self.args[1]))


def set_syntax_version(version, message):
    """Attempt to define the syntax version.

    If it's already defined differently, raise SyntaxVersionError.
    """
    if SyntaxVersion.VERSION is None:
        SyntaxVersion.VERSION = version
        SyntaxVersion.VERSION_REASON = message
    elif SyntaxVersion.VERSION != version:
        raise SyntaxVersionError(version, message)
    if SyntaxVersion.VERSION in DEPRECATED_SYNTAX_WARNING:
        warning = (
            DEPRECATED_SYNTAX_WARNING[SyntaxVersion.VERSION].format(message))
        if warning not in SyntaxVersion.WARNING_MESSAGES:
            sys.stderr.write(warning)
            SyntaxVersion.WARNING_MESSAGES.add(warning)




import sys

if len(sys.argv) > 1:
    from _ordereddict import sorteddict as SD
else:
    from _ordereddict import sorteddict
    class SD(sorteddict):
        def __init__(self, *argc, **kw):
            sorteddict.__init__(self, *argc, **kw)

s = SD({0: 'foo', 2: 'bar'})
print s
print repr(s)
print s.items()
s[1] = 'bla'
print s.items()
